import serial
from serial.tools import list_ports
import time
import math

""" トランジスタ技術 2023年9月号
超安お手軽！ Nucleoのアナログ機能で作る多機能計測ステーション
Nucleo制御用 Python API ライブラリ

<動作確認環境>
    windows 11
    python 3.8.5

<必要ライブラリ>
    pySerial : インストールコマンド pip install pyserial
"""

class DiscoveryBase():
    PID = 0x374B        # STM32 Nucloe USB Product ID
    VID = 0x0483        # USB Vendor ID

    CLOCK_HZ = 80e6     # ペリフェラルクロック周波数[Hz]

    # start_adc() メソッドの引数
    ADC_TRIGGER_RISE = 1
    ADC_TRIGGER_FALL = 2
    ADC_TRIGGER_NONE = 3

    # get_adc_state() メソッドの戻値
    ADC_STATE_STOP = 0
    ADC_STATE_RISE = 1
    ADC_STATE_FALL = 2
    ADC_STATE_SAMPLING = 3
    ADC_STATE_COMPLETE = 4

    # set_gpio() メソッドの引数
    GPIO_L = 0
    GPIO_H = 1
    GPIO_X = 2

    # ADC, DAC波形メモリの要素数
    WAVE_MEMORY_SIZE = 8192

    def __init__(self, device_name=''):
        self._device_name = device_name
        self._device = None

    def _search_device(self)->str:
        device_list = list_ports.comports()
        for i in device_list:
            if i.vid == DiscoveryBase.VID and\
                i.pid == DiscoveryBase.PID:
                return i.device
        return ''

    def open(self)->bool:
        """ Nucleoとの接続確立 """
        if not self._device_name:
            self._device_name = self._search_device()
        if (not self._device) and self._device_name:
            try:
                self._device = serial.Serial(self._device_name, 460800, timeout=0)
                self._device.reset_input_buffer()
                return True
            except:
                self._device = None
                return False
        return False

    def close(self)->bool:
        """ Nucleoとの接続断 """
        if self._device:
            self._device.close()
            self._device = None
        return True

    def is_open(self)->bool:
        """ Nucleoとの接続状態 """
        if self._device:
            return True
        return False

    def set_pwm(self, freq:float, duty:float)->tuple:
        """ PWM出力の条件設定

        PWM出力の条件を設定。PWM ON状態なら即時反映。
        PWMの出力の ON/OFF は、別途 enable_pwm(), disable_pwm() メソッドで制御すること。

        parameters
        -----
        freq: float
            PWM周波数 [Hz]
        duty: float
            Duty比 0-1
        
        returns
        -----
        tuple
            2要素のタプル
            1要素目は、実際に出力されるPWM周波数 [Hz]
            2要素目は、実際に出力されるDuty比 0-1
            失敗すると、空タプルが返る。
        
        電源投入後の初期値
        -----
        freq: 1000000
        duty: 0.50
        """
        if 0 < freq and 0 <= duty <= 1:
            count = round(DiscoveryBase.CLOCK_HZ / freq)
            if count == 0:
                count = 1
                prescale = 1
            elif count > 4294967296:
                prescale = math.ceil(count/4294967296)
                if prescale > 65536:
                    prescale = 65536
                    count = 4294967296
                else:
                    count = round(count / prescale)
            else:
                prescale = 1
            pulse = round(count * duty)
            self._send_command('+p', [prescale-1, count-1, pulse])
            ack = self._get_while_CR()
            if ack and ack[0] == 'Y':
                return (DiscoveryBase.CLOCK_HZ/prescale/count, pulse/count)
        return ()

    def enable_pwm(self)->bool:
        """ PWM出力ON """
        self._send_command('+P', [1])
        ack = self._get_while_CR()
        if ack and ack[0] == 'Y':
            return True
        return False

    def disable_pwm(self)->bool:
        """ PWM出力OFF """
        self._send_command('+P', [0])
        ack = self._get_while_CR()
        if ack and ack[0] == 'Y':
            return True
        return False

    def set_gpio(self, chA:int, chB:int, chC:int, chD:int)->bool:
        """ GPIO出力設定(即時反映)
        
        parameters
        -----
        chA: int
        chB: int
        chC: int
        chD: int
            各GPIOチャンネルの論理レベル
            0: Low, 1: High, それ以外: 現状維持
        
        電源投入後の初期値
        -----
        All Low
        """
        if chA < 0 or chA > 1:
            chA = 2
        if chB < 0 or chB > 1:
            chB = 2
        if chC < 0 or chC > 1:
            chC = 2
        if chD < 0 or chD > 1:
            chD = 2
        self._send_command('+g', [chA, chB, chC, chD])
        ack = self._get_while_CR()
        if ack and ack[0] == 'Y':
            return True
        return False

    def set_dac_sinwave(self, refresh_freq:float, point:int, bias:float, amplitude:float, phase:float=0, repeat_times:int=0)->float:
        """ DAC出力に正弦波を設定

        DAC波形メモリと、波形出力条件の設定を最適化し、所望の正弦波を出力する準備をする。
        DAC OFF時の電圧は、波形メモリ先頭の電圧になる。(位相設定した場合、振幅中点で無いことに注意)
        DAC 出力の ON/OFF は、別途 enable_dac(), disable_dac() メソッドで制御すること。

        parameters
        -----
        refresh_freq: float 推奨値 0.01 - 2e6/point
            出力正弦波の所望周波数 [Hz]
        point: int 2-8192
            1周期の波形ポイント数
        bias: float
        amplitude: float
            正弦波を bias + amplitude*sin() [V] に設定する。DAC OFF 時の電圧は、波形メモリ先頭の電圧とする。
        phase: float
            正弦波の始点の位相 [degree]
        repeat_times: int >= 0
            正弦波の繰返し回数 (0は無限繰返し)

        returns
        -----
        float
            実際に設定された正弦波の周波数 [Hz]
            失敗すると、0 が返る。
        """
        if 2 <= point <= DiscoveryBase.WAVE_MEMORY_SIZE and 0 < refresh_freq and\
           0 <= repeat_times <= 65535 and 0 <= bias+amplitude <= 3.3 and 0 <= bias-amplitude <= 3.3:
            estimate = self._estimate_divfreq(refresh_freq, point)
            if estimate:
                wave_list = self._make_sin_list(bias, amplitude, phase, point)
                self.set_dac_memory(wave_list)
                self.set_dac_raw(0, estimate[0]-1, 0, point-1, repeat_times, wave_list[0])
                return estimate[1]
        return 0

    def set_dac_memory(self, voltage_list:list)->bool:
        """ DAC波形メモリへの転送

        マイコン内蔵のDAC波形メモリに、電圧リストを転送する。
        電圧リストの出力条件は、別途 set_dac() メソッドで設定する。
        DAC ON時は、強制 OFF してから、転送する。

        parameters
        -----
        voltage_list: list
            DAC波形メモリへのデータ転送リスト
            最大8192要素の電圧値とし、電圧範囲は 0 - 3.3[V] とする。(範囲外の電圧は、飽和処理する)
            8192要素未満であれば、不足分は 0 [V] 埋めされる。
        
        電源投入後の初期値
        -----
        不定
        """
        if 0 < len(voltage_list) <= DiscoveryBase.WAVE_MEMORY_SIZE:
            self._send_command('+s', [len(voltage_list)])
            ack = self._get_while_CR()
            if ack and ack[0] == 'Y':
                self._send_binary16([round(i/3.3*4095) if 0 <= i <= 3.3 else (0 if i < 0 else 4095) for i in voltage_list])
                return True
        return False

    def set_dac(self, refresh_time:float, start_index:int, stop_index:int, repeat_times:int=0, off_voltage:float=0)->float:
        """ DAC出力の条件設定

        DAC出力の波形出力条件を設定する。
        あらかじめ set_dac_memory() メソッドで、DAC波形メモリを準備しておく。
        DAC ON時は、一旦出力停止してから、この設定条件を適用して、波形出力再開する。
        DAC OFF時は、off_voltage 電圧値を即時反映する。
        DAC の出力 ON/OFF は、別途 enable_dac(), disable_dac() メソッドで制御すること。

        parameters
        -----
        refresh_time: float 推奨値 (0.5e-6) - 50
            DAC波形メモリの更新間隔時間 [sec]
        start_index: int 0-8191
            DAC波形メモリの、波形始点インデックス
        stop_index: int 0-8191 (>start_index)
            DAC波形メモリの、波形終点インデックス
        repeat_times: int 0-65535
            波形メモリの繰返し回数、0 は無限繰返しを意味する。
            繰返し回数を出力し終えたら、自動的に DAC出力 OFF になる。
        off_voltage: float 0-3.3
            DAC出力OFF時のDAC出力電圧 0-3.3[V]
        
        returns
        -----
        float
            実際に設定された更新間隔時間 [sec]
            失敗すると、0 が返る。
        
        電源投入後の初期値
        -----
        refresh_time: 1e-6
        start_index: 0
        stop_index: 8191
        repeat_times: 0
        off_voltage: 0
        """
        if refresh_time > 0:
            div = round(DiscoveryBase.CLOCK_HZ * refresh_time)
            # DACの最大リフレッシュレートは 2Msps が実用限界(80MHz 40分周)
            if div < 40:
                div = 40
            elif div > 4294967296:
                div = 4294967296
            if self.set_dac_raw(0, div-1, start_index, stop_index, repeat_times, off_voltage):
                return div / DiscoveryBase.CLOCK_HZ
        return 0

    def set_dac_raw(self, prescaler:int, counter:int, start_index:int, stop_index:int,
                    repeat_times:int=0, off_voltage:float=0)->bool:
        """ DAC出力の条件設定 (低水準関数につき、ユーザの利用は推奨しない)

        DAC出力の波形出力条件を設定する。
        あらかじめ set_dac_memory() メソッドで、DAC波形メモリを準備しておく。
        DAC ON時は、一旦停止して、この設定条件を適用してから、波形出力再開する。
        DAC OFF時は、off_voltage 値が即時反映する。
        DAC の出力 ON/OFF は、別途 enable_dac(), disable_dac() メソッドで制御すること。

        parameters
        -----
        prescaler: int 0-65535(16bit)
        counter: int 0-4294967295(32bit)
            DAC波形の更新周波数を、CLOCK_HZ(=80e6)÷(prescaler+1)*(counter+1)[Hz] に設定
            更新周波数 2MHz回数回数下に制限するため、(Prescaler+1)*(Counter+1) >= 40 であること。
        start_index: int 0-8191
            DAC波形メモリの、波形始点インデックス
        stop_index: int 0-8191 (>start_index)
            DAC波形メモリの、波形終点インデックス
        repeat_times: int 0-65535
            波形メモリの繰返し回数、0 は無限繰返しを意味する。
            繰返し回数を出力し終えたら、自動的に DAC出力 OFF になる。
        off_voltage: float 0-3.3
            DAC出力OFF時のDAC出力電圧 0-3.3[V]
        
        電源投入後の初期値
        -----
        prescaler: 0
        counter: 79
        start_index: 0
        stop_index: 8191
        repeat_times: 0
        off_voltage: 0
        """
        if 0 <= prescaler <= 65535 and\
           0 <= counter <= 4294967295 and\
           (prescaler+1)*(counter+1) >= 40 and\
           0 <= start_index < DiscoveryBase.WAVE_MEMORY_SIZE and\
           0 <= stop_index < DiscoveryBase.WAVE_MEMORY_SIZE and\
           start_index < stop_index and\
           0 <= repeat_times <= 65535 and\
           0 <= off_voltage <= 3.3:
            self._send_command('+d', [prescaler, counter, start_index,
                                      stop_index, repeat_times, round(off_voltage/3.3*4095)])
            ack = self._get_while_CR()
            if ack and ack[0] == 'Y':
                return True
        return False

    def enable_dac(self)->bool:
        """ DAC出力ON """
        self._send_command('+D', [1])
        ack = self._get_while_CR()
        if ack and ack[0] == 'Y':
            return True
        return False

    def disable_dac(self)->bool:
        """ DAC出力OFF """
        self._send_command('+D', [0])
        ack = self._get_while_CR()
        if ack and ack[0] == 'Y':
            return True
        return False

    def set_dac_off_voltage(self, off_voltage:float)->bool:
        """ DAC出力OFF 時の出力電圧設定

        DAC出力OFF 時の出力電圧を設定する。
        DAC出力OFF時は即時反映する。

        parameters
        -----
        off_voltage: float 0-3.3
            DAC出力OFF時のDAC出力電圧 0-3.3[V]
        
        電源投入後の初期値
        -----
        0
        """
        if 0 <= off_voltage <= 3.3:
            self._send_command('+e', [round(off_voltage/3.3*4095)])
            ack = self._get_while_CR()
            if ack and ack[0] == 'Y':
                return True
        return False

    def get_monitor(self)->list:
        """ Monitor Ch.A/B/C/D の電圧取得

        returns
        -----
        list
            Ch.A/B/C/D の順で、4要素の電圧値リスト 0-3.3[V]
            失敗すると空リストが返る。
        """
        self._send_command('-m')
        ack = self._get_while_CR()
        if ack and ack[0] == 'Y' and len(ack) == 5:
            return [ack[i+1]/4095*3.3 for i in range(4)]
        return []

    def set_adc_sampling_time(self, sampling_time:float)->float:
        """ ADC Ch.A/B 電圧サンプリングの条件を設定

        ADC のサンプリング間隔時間 [sec] を設定する。ADC 停止中のみ有効。
        本メソッドか、set_adc_sampling_freq() メソッドか、どちらかで ADCサンプリング条件を設定する。

        parameters
        -----
        sampling_time:float 推奨値 (0.25e-6) - 50
            ADC のサンプリング間隔時間 [sec]
        
        returns
        -----
        float
            実際に設定されたサンプリング間隔時間 [sec]
            失敗すると 0 が返る。
        
        電源投入後の初期値
        -----
        sampling_time: 1e-6
        """
        if 0 < sampling_time:
            counter = round(DiscoveryBase.CLOCK_HZ * sampling_time)
            # ADCの最大サンプリングレートは 4MHz (80MHz 20分周)
            if counter < 20:
                counter = 20
                prescaler = 1
            elif counter > 65536:
                prescaler = math.ceil(counter / 65536)
                if prescaler > 65536:
                    prescaler = 65536
                    counter = 65536
                else:
                    counter = round(counter / prescaler)
            else:
                prescaler = 1
            if self.set_adc_raw(prescaler-1, counter-1):
                return prescaler*counter/DiscoveryBase.CLOCK_HZ
        return 0


    def set_adc_sampling_freq(self, sampling_freq:float)->float:
        """ ADC Ch.A/B 電圧サンプリングの条件を設定

        ADC のサンプリング周波数 [Hz] を設定する。ADC 停止中のみ有効。
        本メソッドか、set_adc_sampling_time() メソッドか、どちらかで ADCサンプリング条件を設定する。

        parameters
        -----
        sampling_freq:float 推奨値 0.02 - 4e6
            ADC のサンプリング周波数 [Hz]
        
        returns
        -----
        float
            実際に設定されたサンプリング周波数 [Hz]
            失敗すると 0 が返る。
        
        電源投入後の初期値
        -----
        sampling_freq: 1e6
        """
        if 0 < sampling_freq:
            counter = round(DiscoveryBase.CLOCK_HZ / sampling_freq)
            # ADCの最大サンプリングレートは 4MHz (80MHz 20分周)
            if counter < 20:
                counter = 20
                prescaler = 1
            elif counter > 65536:
                prescaler = math.ceil(counter / 65536)
                if prescaler > 65536:
                    prescaler = 65536
                    counter = 65536
                else:
                    counter = round(counter / prescaler)
            else:
                prescaler = 1
            if self.set_adc_raw(prescaler-1, counter-1):
                return DiscoveryBase.CLOCK_HZ/prescaler/counter
        return 0

    def set_adc_raw(self, prescaler:int, counter:int)->bool:
        """ ADC Ch.A/B 電圧サンプリングの条件を設定 (低水準関数につき、ユーザの利用は推奨しない)

        parameters
        -----
        prescaler: int 0-65535(16bit)
        counter: int 0-65535(16bit)
            ADCのサンプリング周波数を、CLOCK_HZ(=80e6)÷(prescaler+1)*(counter+1)[Hz] に設定
            サンプリング周波数 4MHz 以下に制限するため、(Prescaler+1)*(Counter+1) >= 20 であること。
        
        電源投入後の初期値
        -----
        prescaler: 0
        counter: 79
        """
        if 0 <= prescaler <= 65535 and\
           0 <= counter <= 65535 and\
           (prescaler+1)*(counter+1) >= 20:
            self._send_command('+a', [prescaler, counter])
            ack = self._get_while_CR()
            if ack and ack[0] == 'Y':
                return True
        return False

    def set_trigger(self, trigger_voltage:float)->bool:
        """ トリガ電圧設定(即時反映)

        parameters
        -----
        trigger_voltage: float
            トリガ電圧 0-3.3[V]
        
        電源投入後の初期値
        -----
        0
        """
        if 0 <= trigger_voltage <= 3.3:
            self._send_command('+t', [round(trigger_voltage/3.3*4095)])
            ack = self._get_while_CR()
            if ack and ack[0] == 'Y':
                return True
        return False

    def start_adc(self, trigger:int)->bool:
        """ ADC Ch.A/B 電圧サンプリング開始

        AD変換の開始タイミングを指示して、変換予約をする。変換の状態は、get_adc_state() メソッドで確認する。
        変換状態 ADC_STATE_COMPLETE(=4) であれば、マイコン内の AD変換波形メモリにデータが記録されているので、
        get_adc_memory() メソッドで AD変換サンプリングデータを取得する。

        parameters
        -----
        trigger: int
            ADC_TRIGGER_RISE(=1) TRG入力が、トリガ電圧を超過したタイミングでADC開始
            ADC_TRIGGER_FALL(=2) TRG入力が、トリガ電圧を下回ったタイミングでADC開始
            ADC_TRIGGER_NONE(=3) すぐにADC開始
        """
        if 1 <= trigger <= 3:
            self._send_command('+A', [trigger])
            ack = self._get_while_CR()
            if ack and ack[0] == 'Y':
                return True
        return False

    def stop_adc(self)->bool:
        """ ADC Ch.A/B 入力電圧サンプリング強制停止 """
        self._send_command('+A', [0])
        ack = self._get_while_CR()
        if ack and ack[0] == 'Y':
            return True
        return False

    def get_adc_state(self)->int:
        """ ADC Ch.A/B 電圧サンプリング状態取得

        returns
        -----
        ADC_STATE_STOP(=0) サンプリング停止状態
        ADC_STATE_RISE(=1) 立上りトリガ待ち状態
        ADC_STATE_FALL(=2) 立下りトリガ待ち状態
        ADC_STATE_SAMPLING(=3) サンプリング中
        ADC_STATE_COMPLETE(=4) サンプリング完了
            サンプリング終了すると、サンプリングデータはマイコン内の AD変換波形メモリに保存されている。
            次にサンプリング開始するまでメモリは保持される。
        
        電源投入後の初期値
        -----
        ADC_STATE_STOP(=0)
        """
        self._send_command('-A')
        ack = self._get_while_CR()
        if ack and ack[0] == 'Y':
            return ack[1]
        return 0

    def get_adc_memory(self)->list:
        """ ADC Ch.A/B 電圧サンプリングデータ取得

        マイコン内部に保存された AD変換サンプリングデータを取得する。
        adc_start() を呼び出し、get_adc_state() による状態が ADC_STATE_COMPLETE であれば有意なデータである。

        returns
        -----
        2要素のリスト
        第1要素は Ch.A の 8192点の電圧リスト 0-3.3 [V]
        第2要素は Ch.B の 8192点の電圧リスト 0-3.3 [V]
        取得失敗時は空リスト。

        例) [[1.23, 1.25, 1.32, 1.40, ... ], [2.1, 1.9, 1.0, 2.2, 1.2 ... ]]
             <---- Ch.A の時系列データ ---->  <---- Ch.B の時系列データ ---->
                     8192点                              8192点
        """
        self._send_command('-r')
        ack = self._get_binary16(DiscoveryBase.WAVE_MEMORY_SIZE*2)
        if ack:
            return [ [i/4095*3.3 for i in ack[:DiscoveryBase.WAVE_MEMORY_SIZE]],
                     [i/4095*3.3 for i in ack[DiscoveryBase.WAVE_MEMORY_SIZE:]] ]
        else:
            return []

    def set_logger_sampling_time(self, sampling_time:float)->float:
        """ Monitor電圧ロギングモード(システムモード SYSTEM_MODE_LOG) におけるログデータ条件を設定

        ロギングモードのサンプリング間隔時間 [sec] を設定する。
        本メソッドか、set_logger_sampling_freq() メソッドか、どちらかでサンプリング条件を設定する。
        本メソッドは条件設定だけを行う。Monitor電圧ロギングモードに移行するには、別途コマンド "+L\r" が必要。
        プロット終了時は、コマンド "+F\r" を送る。

        parameters
        -----
        sampling_time:float
            ロギングモードのサンプリング間隔時間 [sec] 0.01以上
        
        returns
        -----
        float
            実際に設定されたサンプリング間隔時間 [sec]
            失敗すると 0 が返る。
        
        電源投入後の初期値
        -----
        sampling_time: 0.01
        """
        if 0.01 <= sampling_time:
            counter = round(DiscoveryBase.CLOCK_HZ * sampling_time)
            if counter == 0:
                counter = 1
                prescaler = 1
            elif counter > 65536:
                prescaler = math.ceil(counter / 65536)
                if prescaler > 65536:
                    prescaler = 65536
                    counter = 65536
                else:
                    counter = round(counter / prescaler)
                    if prescaler * counter < 800000:
                        counter += 1
            else:
                prescaler = 1
            if self.set_logger_raw(prescaler-1, counter-1):
                return prescaler*counter/DiscoveryBase.CLOCK_HZ
        return 0

    def set_logger_sampling_freq(self, sampling_freq:float)->float:
        """ Monitor電圧ロギングモード(システムモード SYSTEM_MODE_LOG) におけるログデータ条件を設定

        ロギングモードのサンプリング周波数 [Hz] を設定する。
        本メソッドか、set_logger_sampling_time() メソッドか、どちらかでサンプリング条件を設定する。
        本メソッドは条件設定だけを行う。Monitor電圧ロギングモードに移行するには、別途コマンド "+L\r" が必要。
        プロット終了時は、コマンド "+F\r" を送る。

        parameters
        -----
        sampling_freq:float
            ADC のサンプリング周波数 [Hz] 100以下
        
        returns
        -----
        float
            実際に設定されたサンプリング周波数 [Hz]
            失敗すると 0 が返る。
        
        電源投入後の初期値
        -----
        sampling_freq: 100
        """
        if 0 < sampling_freq <= 100:
            counter = round(DiscoveryBase.CLOCK_HZ / sampling_freq)
            if counter == 0:
                counter = 1
                prescaler = 1
            elif counter > 65536:
                prescaler = math.ceil(counter / 65536)
                if prescaler > 65536:
                    prescaler = 65536
                    counter = 65536
                else:
                    counter = round(counter / prescaler)
                    if prescaler * counter < 800000:
                        counter += 1
            else:
                prescaler = 1
            if self.set_logger_raw(prescaler-1, counter-1):
                return DiscoveryBase.CLOCK_HZ/prescaler/counter
        return 0

    def set_logger_raw(self, prescaler:int, counter:int)->bool:
        """ Monitor電圧ロギングモード(システムモード SYSTEM_MODE_LOG) におけるログデータ条件を設定

        本メソッドは条件設定だけを行う。Monitor電圧ロギングモードに移行するには、別途コマンド "+L\r" が必要。
        一旦本クラスインスタンスの close() メソッドで、通信切断した後、シリアルプロッタソフトウェアを起動して
        Nucleo と COM接続し、シリアルプロッタ内でシリアルコマンド送信 "+L\r" を送り、プロット開始する。
        プロット終了時は、コマンド "+F\r" を送る。

        parameters
        -----
        prescaler: int 0-65535(16bit)
        counter: int 0-65535(16bit)
            ログデータの更新周波数を、CLOCK_HZ(=80e6)÷(prescaler+1)*(counter+1)[Hz] に設定
            ログ周期100Hz 以下に制限するため、(Prescaler+1)*(Counter+1) >= 800000 であること。
        """
        if 0 <= prescaler <= 65535 and\
           0 <= counter <= 65535 and\
           (prescaler+1)*(counter+1) >= 800000:
            self._send_command('+l', [prescaler, counter])
            ack = self._get_while_CR()
            if ack and ack[0] == 'Y':
                return True
        return False

    # ------------------------------------------------------
    # 以降は内部呼び出し用途につき、外部から直接呼び出さないこと
    # ------------------------------------------------------

    def _send_command(self, header:str, params:list=[])->None:
        """ コマンドを Nucleo Discovery へ送信

        送信は、header 文字に加えて、params 数値リストを8桁Hex文字列
        で連ねて、最後に終端文字 CR (\r) 付加して、PCへ送信する。
        
        parameters
        -----
        header: str
            コマンドヘッダ文字 (例: "+D")
        params: list
            数値のリスト
        """
        if self._device:
            message = header
            for i in params:
                if i < 0:
                    i = 0
                elif i > 0xFFFFFFFF:
                    i = 0xFFFFFFFF
                message += '{:08X}'.format(i)
            message += '\r'
            self._device.reset_input_buffer()
            self._device.write(message.encode())

    def _print_command(self, header:str, params:list=[])->str:
        """ コマンドをテキスト化

        parameters
        -----
        header: str
            コマンドヘッダ文字 (例: "+D")
        params: list
            数値のリスト
        """
        message = header
        for i in params:
            if i < 0:
                i = 0
            elif i > 0xFFFFFFFF:
                i = 0xFFFFFFFF
            message += '{:08X}'.format(i)
        message += '\r'
        return message

    def _get_while_CR(self, timeout:float=0.1)->list:
        """ コマンドの戻りを受信

        受信文字列は終端文字('\r') まで検出する。結果をリスト取得する。

        parameters
        -----
        timeout: float, default 1.0
            受信文字列の受け取りタイムアウト時間[sec]

        returns
        -----
        コマンド戻値のリスト、第一要素は 'Y'/ 'N' の文字、第二要素以降に数値の戻値が格納される。
        タイムアウト時は空リスト。
        """
        start_time = time.time()
        ack_list = []
        if self._device:
            ack = b''
            while time.time() - start_time < timeout:
                ack += self._device.read_all()
                if ack and ack[-1] == 13:
                    break
            else:
                return []
            ack = ack.decode()
            ack_len = len(ack)
            if (ack_len-2) % 8 == 0:
                if (ack[0] == 'Y' or ack[0] == 'N') and ack[-1] == '\r':
                    ack_list.append(ack[0])
                    for i in range((ack_len-2)//8):
                        ack_list.append(int(ack[i*8+1: i*8+9], 16))
        return ack_list

    def _get_binary16(self, points:int, timeout:float=3)->list:
        """ バイナリで直接コマンド受信

        バイナリデータを直接受信し、2byte littleエンディアンの無符号数値とみなして、
        数値リストで取得する。

        parameters
        -----
        points: int
            2byte数値の総受信個数
        timeout: float, default 10.0
            受信文字列の受け取りタイムアウト時間[sec]
        
        returns
        -----
        list
            受信されたバイナリを数値変換して得られる、数値リスト。
            タイムアウト時は空リスト。
        """
        start_time = time.time()
        byte_len = points*2
        ack_list = []
        if self._device:
            ack = b''
            while time.time() - start_time < timeout:
                ack += self._device.read_all()
                if len(ack) >= byte_len:
                    break
            else:
                return []
            for i in range(points):
                ack_list.append(int.from_bytes(ack[i*2: (i+1)*2], byteorder='little'))
        return ack_list

    def _send_binary16(self, send_list:list)->None:
        """ バイナリで直接コマンド送信

        バイナリデータを直接送信する。引数のリスト数値を2byte 無符号数値とみなして、
        littleエンディアン形式で送信する。

        parameters
        -----
        send_list: list
            無符号16bit で表現できる数の、数値リスト
        """
        if self._device:
            for i in send_list:
                if i < 0:
                    i = 0
                elif i > 0xFFFF:
                    i = 0xFFFF
                self._device.write(i.to_bytes(2, byteorder='little'))

    def _make_sin_list(self, bias:float, amplitude:float, phase:float, point:int)->list:
        """ 正弦波のリストを生成する

        parameters
        -----
        bias: float
            オフセット電圧[V]
        amplitude: float
            振幅[V]
        phase: float
            位相[degree]
            出力電圧 [V] = bias + amplitude*sin(2*PI*n/point+2*PI*phase/360)
            になる正弦波を、DA変換バイナリ値に変換して、リスト出力する。n: 0 - (point-1)
        point: int
            1周期の点数 (2-8192)
        
        returns
        -----
        list
            要素数point の電圧値リスト 0-3.3 [V]
            不当な引数値の場合は空リスト。
        """
        if 2 <= point <= DiscoveryBase.WAVE_MEMORY_SIZE and 0 <= bias+amplitude <= 3.3 and 0 <= bias-amplitude <= 3.3:
            return [bias + amplitude*math.sin(2*math.pi*(i/point+phase/360)) for i in range(point)]
        else:
            return []

    def _estimate_divfreq(self, desired_freq:float, point:int)->tuple:
        """ DAC波形メモリの1周期点数と、目標とする1周期周波数から、設定可能な近い1周期周波数を見積もる

        (例) 目標周波数 23kHz, 1周期ポイント数 128 を指定した場合
                ペリフェラルの励振クロックは80MHzなので 27分周で128点周期となる周波数 23.148148 [kHz] が一番近い
                return => (27, 23148.148)
        
        parameters
        -----
        desired_freq: float 目標周波数[Hz]
        point: int 1周期のポイント数 ( > 1)

        returns
        -----
        tuple
            要素数2 のタプル
            第1要素は、ペリフェラルクロック周波数に対する分周数 ( > 0)
            第2要素は、目標周波数に最も近い設定可能周波数 [Hz]
            計算不可の場合は、空タプル
        """
        if point < 2:
            return ()
        div = round(DiscoveryBase.CLOCK_HZ/desired_freq/point)
        # DACの最大リフレッシュレートは 2Msps が実用限界(80MHz 40分周)
        if div < 40:
            div = 40
        elif div > 0x100000000:
            div = 0x100000000
        return (div, DiscoveryBase.CLOCK_HZ/div/point)
