
#ifndef INC_COMMAND_H_
#define INC_COMMAND_H_

#include "stdint.h"

/* [コマンド受信形式]
 * 		設定系のコマンド
 * 			先頭は '+'、その後英字1文字、その後複数個のコマンド引数で構成される。各部を仕切る文字は無く、単純に連結する。
 * 			終端文字('\r') を最後に付加。コマンド引数は、8桁のHex文字列(大文字小文字問わず)。
 * 				"+ 英字   8桁Hex文字列   8桁Hex文字列 ...  終端文字(\r)" => 例  "+p0000000A000003D20000000F\r"
 *
 * 		取得系のコマンド
 * 			先頭は '-'、その後英字1文字、終端文字('\r') を最後に付加。
 * 				"- 英字   終端文字(\r)" => "+p\r"
 *
 *
 * [コマンド戻値形式]
 * 		コマンドが正しく処理されたら、先頭1文字は 'Y' として、複数個のコマンド戻値が続く。
 * 		複数個のコマンド戻値は、1個あたり8桁Hex文字列で、終端文字は CR('\r')。
 * 		(例) 3個のコマンド戻値  "Y000012000000A00300004DA5\r"  =>  0x1200, 0xA003, 0x4DA5
 *
 * 		不法なコマンドの場合は、先頭1文字は 'N' として、終端文字('\r') を最後に付加。"N\r"
 *
 * 		コマンド "+F\r" / "+L\r" は、戻値は返さない。
 * 		システムモード SYSTEM_MODE_LOG の場合も、戻値は返さない。
 *
 *
 * [各コマンド]
 * 		-l
 * 			<意味>
 * 			Monitor電圧ロギングモード(システムモード SYSTEM_MODE_LOG) におけるログデータ条件を取得
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			2個の変数戻値
 * 			ログ間隔周期となる80MHz タイマTIM4 の Prescaler(16bit), Counter(16bit) 値
 * 			ログ間隔周期 = 80÷(Prescaler+1)÷(Counter+1) [MHz]
 *
 *		+l
 *			<意味>
 * 			Monitor電圧ロギングモード(システムモード SYSTEM_MODE_LOG) におけるログデータ条件を設定
 * 			<コマンド引数>
 * 			2個のパラメータ
 * 			ログ間隔周期となる80MHz タイマTIM4 の Prescaler(16bit), Counter(16bit) 値
 * 			ログ間隔周期 = 80÷(Prescaler+1)÷(Counter+1) [MHz]
 * 			ログ間隔周期100Hz 以下に制限するため、(Prescaler+1)*(Counter+1) >= 800000 であること。
 * 			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			39999, 19 (100Hzに相当)
 *
 * 		-p
 * 			<意味>
 * 			PWM出力の条件を取得
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			3個の変数戻値
 * 			80MHz タイマTIM2-CH3のPWM mode1の Prescaler(16bit), Counter(32bit), Pulse(32bit)
 * 			1クロック 80÷(Prescaler+1) [MHz] として、PWM1周期 (Counter+1) クロック、ON時間 (Pulse) クロックとする、PWM波形
 *
 *		+p
 *			<意味>
 * 			PWM出力の条件を設定
 * 			<コマンド引数>
 * 			3個のパラメータ
 * 			80MHz タイマTIM2-CH3のPWM mode1の Prescaler(16bit), Counter(32bit), Pulse(32bit)
 * 			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0, 79, 40 (1MHz Duty50% に相当)
 *
 * 		-P
 * 			<意味>
 * 			PWM出力の状態を取得
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			1個の変数戻値
 * 			0: OFF / 1: ON
 *
 *		+P
 *			<意味>
 * 			PWM出力の状態を設定
 * 			<コマンド引数>
 * 			1個のパラメータ
 * 			0: OFF / 1: ON
 * 			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0 (OFFに相当)
 *
 * 		-t
 * 			<意味>
 * 			トリガ閾値電圧を取得
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			1個の変数戻値
 * 			DAC値バイナリ値0-4095 (0-3.3V に対応)
 *
 *		+t
 * 			<意味>
 * 			トリガ閾値電圧を設定し、即時反映。
 * 			<コマンド引数>
 * 			1個のパラメータ
 * 			DAC値バイナリ値0-4095 (0-3.3V に対応)
 * 			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0
 *
 * 		-d
 * 			<意味>
 * 			DAC Ch.A 出力の条件を取得
 * 			一定周期毎に、DAC波形メモリに従った電圧を出力する
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			6個の変数戻値
 * 			80MHz タイマ Tim5 の Prescaler(16bit), Counter(32bit), Start index, Stop index, Repeat times(16bit), OFF value(12bit)
 *
 * 			一定周期を作り出す 80MHz タイマ TIM5 の Prescaler(16bit), Counter(32bit)
 * 			一定周期の周波数 = 80÷(Prescaler+1)÷(Counter+1) [MHz]
 * 			DAC波形メモリの要素番号 Start index ～ Stop index までを繰返し出力する ( < DAC_MEMORY_SIZE )
 * 			波形の繰返し回数 Repeat times(16bit) (但し 0 は無限ループを意味する)
 * 			繰返し回数を出力し終えたら、自動的にOFFする。
 * 			DAC出力OFF時のDACバイナリ値 OFF value(12bit) 0-4095 (0-3.3V に対応)
 *
 *		+d
 *			<意味>
 * 			DAC Ch.A 出力の条件を設定。
 * 			DAC出力ON状態なら、一時停止してから、再出力する。
 * 			<コマンド引数>
 * 			6個のパラメータ
 *			80MHz タイマ Tim5 の Prescaler(16bit), Counter(32bit),  Start index, Stop index, Repeat times(16bit), OFF value(12bit)
 *			一定周期の周波数2MHz 以下に制限するため、(Prescaler+1)*(Counter+1) >= 40 であること。
 *			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0, 79, 0, 8191, 0, 0
 *
 *		-e
 *			<意味>
 * 			DAC Ch.A 出力 OFF 時の DACバイナリ値のみ取得。
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			1個の変数戻値
 * 			OFF value(12bit) 0-4095 (0-3.3V に対応)
 *
 *		+e
 *			<意味>
 * 			DAC Ch.A 出力 OFF 時の DACバイナリ値のみ設定。
 * 			DAC出力OFF状態なら即時反映
 * 			<コマンド引数>
 * 			1個のパラメータ
 *			OFF value(12bit) 0-4095 (0-3.3V に対応)
 *			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0
 *
 *		+s
 *			<意味>
 * 			DAC Ch.A 出力のDAC波形メモリへのデータ転送開始。
 * 			DAC出力ON状態なら、DAC出力強制 OFFにする。
 * 			<コマンド引数>
 * 			1個のパラメータ
 * 			転送するデータ点数( <= DAC_MEMORY_SIZE )
 * 			DAC波形メモリは、DAC_MEMORY_SIZE個分のデータ点数を有している。
 * 			転送したデータ点以降は、ゼロ埋めされる。
 * 			<戻値>
 * 			0個の変数戻値
 * 			正常受理されると、戻値を返してから、モード SYSTEM_MODE_DIRECT に移行する。
 * 			以降バイナリ通信となり、転送データ点数を受け取り後、モード SYSTEM_MODE_FREE に戻る。
 * 			転送データは、1点2byte のリトルエンディアン形式とし、終端文字は付けない。
 *
 * 		-D
 * 			<意味>
 * 			DAC Ch.A 出力の状態を取得
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			1個の変数戻値
 * 			0: OFF / 1: ON
 *
 *		+D
 *			<意味>
 * 			DAC Ch.A 出力の状態を設定
 * 			<コマンド引数>
 * 			1個のパラメータ
 * 			0: OFF / 1: ON
 * 			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0
 *
 * 		-a
 * 			<意味>
 * 			ADC Ch.A/B 電圧サンプリングの条件を取得
 * 			一定周期毎にADC電圧サンプリングし、ADC波形メモリ転送する。
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			2個の変数戻値
 * 			一定周期を作り出す 80MHz タイマ TIM3 の Prescaler(16bit), Counter(16bit)
 * 			一定周期の周波数 = 80÷(Prescaler+1)÷(Counter+1) [MHz]
 *
 *		+a
 *			<意味>
 * 			ADC Ch.A/B 電圧サンプリングの条件を設定。
 * 			ADC開始中 (トリガ待ち/サンプリング中) は無視される。
 * 			<コマンド引数>
 * 			2個のパラメータ
 * 			一定周期を作り出す 80MHz タイマ TIM3 の Prescaler(16bit), Counter(16bit)
 * 			一定周期の周波数 = 80÷(Prescaler+1)÷(Counter+1) [MHz]
 * 			一定周期の周波数4MHz 以下に制限するため、(Prescaler+1)*(Counter+1) >= 20 であること。
 *			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0, 79
 *
 * 		-r
 * 			<意味>
 * 			ADC Ch.A/B 電圧サンプリングデータ(ADC波形メモリ) を取得
 * 			前回取得されたADC電圧サンプリングデータは、次にサンプリングされるまで、ADC波形メモリに保存され続ける。
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値 (特殊形式)>
 * 			2chの波形メモリが、Ch.A , Ch.B の順番に返ってくる。
 * 			1ch当たりデータ点数 ADC_MEMORY_SIZE 点とし、1点2byteのリトルエンディアン形式バイナリデータ。
 * 			終端文字は付かない。
 * 			1点2byteのバイナリ値は 0-4095 であり、0-3.3V に対応する。要約すると、以下の通りとなる。
 *
 * 			ADC Ch.A の 0点目データLowerバイト   0点目データHigherバイト  1点目データLowerバイト   1点目データHigherバイト  ...
 * 			(ADC_MEMORY_SIZE-1)点目データ、(ADC_MEMORY_SIZE-1) 点目データHigherバイト ...
 * 			ADC Ch.B の 0点目データLowerバイト   0点目データHigherバイト  1点目データLowerバイト   1点目データHigherバイト  ...
 * 			(ADC_MEMORY_SIZE-1)点目データ、(ADC_MEMORY_SIZE-1) 点目データHigherバイト
 *
 * 		-A
 * 			<意味>
 * 			ADC Ch.A/B 電圧サンプリングの状態を取得
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			1個の変数戻値
 * 			0: 停止状態で、ADC波形メモリの内容は不定
 * 			1: 立上りエッジトリガが掛かっていて、トリガ待ち
 * 			2: 立下りエッジトリガが掛かっていて、トリガ待ち
 * 			3: サンプリング中
 * 			4: サンプリングが終了しており、ADC波形メモリに有意なデータが存在する
 *
 * 		+A
 * 			<意味>
 * 			ADC Ch.A/B 電圧サンプリングの状態を設定
 * 			<コマンド引数>
 * 			1個のパラメータ
 * 			0: 強制停止にする(ADC波形メモリは不変で保持されている)
 * 			1: 立上りエッジトリガが掛かり、トリガ待ちにする
 * 			2: 立下りエッジトリガが掛かり、トリガ待ちにする
 * 			3: 即サンプリング開始する
 * 			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0
 *
 * 		-g
 * 			<意味>
 * 			GPIO出力の論理状態を取得
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			4個の変数戻値、GPIO Ch.A, B, C, D の順番で
 * 			0: Low, 1: High
 *
 * 		+g
 * 			<意味>
 * 			GPIO出力の論理状態を設定、即座に出力値が切り替わる。
 * 			<コマンド引数>
 * 			4個のパラメータ
 * 			GPIO Ch.A, B, C, D の順番で、0: Low, 1: High, 2: 現在値そのまま
 * 			<戻値>
 * 			0個の変数戻値
 * 			<初期値>
 * 			0, 0, 0, 0
 *
 *		-m
 *			<意味>
 * 			MONITOR Ch.A/B/C/D の電圧取得
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			4個の変数戻値
 * 			MONITOR Ch.A, B, C, D の順番で、ADCバイナリ値 0-4095 (0-3.3V に対応)
 *
 *		+F
 *			<意味>
 * 			システムモードを SYSTEM_MODE_FREE に移行させる。(初期はこのモード)
 * 			現在のシステムモードが SYSTEM_MODE_LOG,  SYSTEM_MODE_FREEの場合にのみ有効。
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			何も応答しない
 *
 *		+L
 *			<意味>
 * 			システムモードを SYSTEM_MODE_LOG に移行させる。
 * 			このモードに移行すると、コマンド "+F\r" 以外を無視し、かつ応答しない。
 * 			<コマンド引数>
 * 			0個のパラメータ
 * 			<戻値>
 * 			何も応答しない
 *
 *
 * [システムモード]
 * 		1. SYSTEM_MODE_FREE
 * 			初期状態のモードで、全てのコマンドを受け付ける。コマンド毎に、アクションをし、戻値を返す。
 * 			戻値を返してから、次のコマンドを受け付ける。
 *
 * 		2. SYSTEM_MODE_LOG
 * 			MONITOR Ch.A/ B/ C/ D の電圧値を、一定周期毎にサンプリングする。
 * 			1サンプリング毎に、UART出力する。出力形式は、4つの10進文字をカンマ区切りで、終端に改行コード ("\r\n") 。
 * 			一般的にシリアルプロッタと呼ばれるフリーソフトウェアで、リアルタイムグラフ表示するに適した出力形式となっている。
 * 			このモードになると、コマンド "+F\r" 以外を受け付けない。コマンド戻値も返さない。
 * 			例)   "121,309,1022,3\r\n129,300,1024,5\r\n131,301,1030,8\r\n...."
 *
 *		3. SYSTEM_MODE_DIRECT
 *			コマンド "s\r" による、DAC波形メモリへのデータ転送状態。指定サイズのバイナリデータを受けるまで継続する。
 *			指定サイズのバイナリデータを受け取った後は、SYSTEM_MODE_FREE に自動復帰する。
 */


#define COMMAND_FALSE	(0)
#define COMMAND_TRUE		(1)

#define COMMAND_TERMINAL ('\r')

// コマンド引数の最大要素数
#define COMMAND_PARAMETERS	(10)
// 戻値の最大要素数
#define COMMAND_RETURNS	(10)
// コマンド構造体
typedef struct {
	uint8_t effective; // 0: 不当なコマンド, 1: 正当なコマンド
	uint8_t command_header;	// +, -
	uint8_t command_type;		// A-Z, a-z
	uint32_t params[COMMAND_PARAMETERS];
} CommandTypeDef;

// Command header
#define COMMAND_SET_HEADER 	('+')
#define COMMAND_GET_HEADER 	('-')

// Command type
#define COMMAND_MODE_FREE		('F')
#define COMMAND_MODE_LOG		('L')

#define COMMAND_LOG_PARAM		('l')
#define COMMAND_PWM_PARAM	('p')
#define COMMAND_PWM_STATE		('P')
#define COMMAND_TRG_PARAM		('t')
#define COMMAND_DAC_PARAM		('d')
#define COMMAND_DAC_OFFBINARY	('e')
#define COMMAND_DAC_MEMORY	('s')
#define COMMAND_DAC_STATE		('D')
#define COMMAND_ADC_PARAM		('a')
#define COMMAND_ADC_MEMORY	('r')
#define COMMAND_ADC_STATE		('A')
#define COMMAND_GPIO_PARAM		('g')
#define COMMAND_MON_VALUE		('m')
// <TODO> コマンドを追加する場合は、ここにコマンド種別文字を定義
// #define COMMAND_XXX  ('x')

// System mode
//	全てのコマンドを受け付けるフリーモード (初期状態)
#define SYSTEM_MODE_FREE		(0)
//	DACメモリへのデータ転送モード
//		コマンド "+s\r" により遷移。指定サイズのバイナリデータを受信し、指定メモリに格納する。
//		指定サイズ受信後には、自動的にフリーモードに移行する。
#define SYSTEM_MODE_DIRECT	(1)
// Monitor電圧ロギングモード
//		コマンド "+L\r" により遷移。Monitor端子4ch(Ch.A/B/C/D) の AD変換データを、指定周期で出力し続ける。
//		このモードに遷移した後は、コマンド "+F\r" (フリーモードに遷移) 以外のコマンドを受け付けない
#define SYSTEM_MODE_LOG		(8)

// Command state
//		各ペリフェラルの動作状態
#define PWM_STATE_OFF		(0)
#define PWM_STATE_ON		(1)
//		DAC が OFF で、指定の OFF電圧値に設定されている
#define DAC_STATE_OFF		(0)
//		DAC が ON で、指定メモリ領域の電圧値を指定周期で更新、指定メモリ領域のデータを、指定回数繰返し掃引する
#define DAC_STATE_ON		(1)
//		ADC が OFF
#define ADC_STATE_OFF		(0)
//		ADC が立上りエッジトリガ待ち
#define ADC_STATE_WAIT_RISE	(1)
//		ADC が立下りエッジトリガ待ち
#define ADC_STATE_WAIT_FALL	(2)
// 	ADC がサンプリング中
#define ADC_STATE_SAMPLING	(3)
//		ADC がサンプリング終了している (有意なデータを取り出せる状態)
#define ADC_STATE_COMPLETE	(4)
#define GPIO_L	(0)
#define GPIO_H	(1)
//		GPIO出力設定時にのみ有効な、現状の論理レベルを継続する状態
#define GPIO_X	(2)

// ADC/DAC波形メモリの要素数(1要素2byte)
#define DAC_MEMORY_SIZE		(8192)
#define ADC_MEMORY_SIZE		(8192)
// Monitor チャンネル数
#define MONITOR_PINS				(4)
// GPIO チャンネル数
#define GPIO_PINS						(4)

// 文字列から、コマンドを解析し、コマンド構造体に変換する。
uint8_t command_analysis(const uint8_t *command_str, uint8_t command_size, CommandTypeDef *command_def);

#endif /* INC_COMMAND_H_ */
