
#include "action.h"
#include "uart_rx.h"
#include "stm32l4xx_hal.h"
#include <string.h>

static ADC_HandleTypeDef *gpAdcChA = 0;
static ADC_HandleTypeDef *gpAdcChB;
static ADC_HandleTypeDef *gpMonitor;
static DMA_HandleTypeDef *gpDmaAdcChA;
static DMA_HandleTypeDef *gpDmaAdcChB;
static DMA_HandleTypeDef *gpDmaAdcMonitor;
static COMP_HandleTypeDef *gpComp;
static DAC_HandleTypeDef *gpDac;
static DMA_HandleTypeDef *gpDmaDac;
static OPAMP_HandleTypeDef *gpOpamp;
static TIM_HandleTypeDef *gpTimPwm;
static TIM_HandleTypeDef *gpTimAdc;
static TIM_HandleTypeDef *gpTimMonitor;
static TIM_HandleTypeDef *gpTimDac;
static UART_HandleTypeDef *gpUart;

static uint8_t system_mode;
volatile static uint16_t adc_memory1[ADC_MEMORY_SIZE];
volatile static uint16_t adc_memory2[ADC_MEMORY_SIZE];
volatile static uint16_t dac_memory[DAC_MEMORY_SIZE];
volatile static uint16_t monitor_memory[MONITOR_PINS];
volatile static uint8_t flg_monitor;
volatile static uint8_t flg_adc_error=0, flg_dac_error=0;

// ペリフェラルの状態管理用の構造体
static struct {
	uint16_t prescaler;
	uint16_t counter;
} log_params;
static struct {
	uint8_t state;
	uint16_t prescaler;
	uint32_t counter;
	uint32_t pulse;
} pwm_params;
static struct {
	uint16_t binary;
} trg_params;
static struct {
	uint8_t state;
	uint16_t prescaler;
	uint32_t counter;
	uint16_t start_index;
	uint16_t stop_index;
	uint16_t repeat_times;
	uint16_t off_binary;
	// running state
	uint16_t now_times;
} dac_params;
static struct {
	uint16_t prescaler;
	uint16_t counter;
	uint8_t state;
} adc_params;
static struct {
	uint8_t state[GPIO_PINS];
} gpio_params;

static void set_gpio(const uint8_t *parameter);
static void set_led(uint8_t level);
static void uart_hex_send(const uint32_t *parameter, uint8_t size);
static void uart_dec_send(const uint16_t *parameter, uint8_t size);
static void uart_bin_send(const uint16_t *parameter, uint16_t size);
static void disable_dac_trigger(void);
static void enable_dac_trigger(void);
static void rewind_tim(TIM_HandleTypeDef *pTim);

void HAL_ADC_ConvCpltCallback(ADC_HandleTypeDef *hadc);
void HAL_ADC_ErrorCallback(ADC_HandleTypeDef *hadc);
void HAL_DACEx_ConvCpltCallbackCh2(DAC_HandleTypeDef *hdac);
void HAL_DACEx_ErrorCallbackCh2(DAC_HandleTypeDef *hdac);
void HAL_DACEx_DMAUnderrunCallbackCh2(DAC_HandleTypeDef *hdac);
void HAL_COMP_TriggerCallback(COMP_HandleTypeDef *hcomp);
void HAL_TIM_PeriodElapsedCallback(TIM_HandleTypeDef *htim);

/* ペリフェラル管理、内部状態管理を司る action.c の初期化
 *
 * ペリフェラルの操作、状態管理は全て action.c の内部で行う。
 * main.c で自動生成された、ペリフェラルのハンドラを受け取り、action.c 内で使用する。
 *
 * [引数]
 * 		各ペリフェラルの HandleTypeDef
 */
void action_init(ADC_HandleTypeDef *pAdcChA,			// ADC Ch.A <= hadc1
						   ADC_HandleTypeDef *pAdcChB,			// ADC Ch.B <= hadc3
						   ADC_HandleTypeDef *pMonitor,			// MONITOR Ch.A/ B/ C/ D <= hadc2
						   DMA_HandleTypeDef *pDmaAdcChA,	// DMA for ADC Ch.A <= hdma_adc1
						   DMA_HandleTypeDef *pDmaAdcChB,	// DMA for ADC Ch.B <= hdma_adc3
						   DMA_HandleTypeDef *pDmaAdcMonitor,  // DMA for MONITOR Ch.A/ B/ C/ D <= hdma_adc2
						   COMP_HandleTypeDef *pComp,			// COMP <= hcomp2
						   DAC_HandleTypeDef *pDac,					// DAC TRG, DAC Ch.A <= hdac1
						   DMA_HandleTypeDef *pDmaDac	,		// DMA for DAC Ch.A <= hdma_dac_ch2
						   OPAMP_HandleTypeDef *pOpamp,		// OPAMP <= hopamp2
						   TIM_HandleTypeDef *pTimPwm, 			// PWM Timer <= htim2
						   TIM_HandleTypeDef *pTimAdc,			// ADC Timer <= htim3
						   TIM_HandleTypeDef *pTimMonitor,		// LOG Timer <= htim4
						   TIM_HandleTypeDef *pTimDac,			// DAC Timer <= htim5
						   UART_HandleTypeDef *pUart  				// UART <= huart2
						   ){
	uint8_t i;

	if(! gpAdcChA){
		// ハンドラ(アドレス) を action.c 静的グローバル変数にコピー
		gpAdcChA = pAdcChA;
		gpAdcChB = pAdcChB;
		gpMonitor = pMonitor;
		gpDmaAdcChA = pDmaAdcChA;
		gpDmaAdcChB = pDmaAdcChB;
		gpDmaAdcMonitor = pDmaAdcMonitor;
		gpComp = pComp;
		gpDac = pDac;
		gpDmaDac = pDmaDac;
		gpOpamp = pOpamp;
		gpTimPwm = pTimPwm;
		gpTimAdc = pTimAdc;
		gpTimMonitor = pTimMonitor;
		gpTimDac = pTimDac;
		gpUart = pUart;

		// 初期システムモード
		system_mode = SYSTEM_MODE_FREE;

		// ペリフェラル状態管理構造体の初期化と、ペリフェラル初期化
		log_params.prescaler = (uint16_t) gpTimMonitor->Instance->PSC;
		log_params.counter = (uint16_t) gpTimMonitor->Instance->ARR;
		pwm_params.prescaler = (uint16_t) gpTimPwm->Instance->PSC;
		pwm_params.counter = gpTimPwm->Instance->ARR;
		pwm_params.pulse = gpTimPwm->Instance->CCR3;
		pwm_params.state = PWM_STATE_OFF;
		trg_params.binary = 0;
		HAL_DAC_SetValue(gpDac, DAC_CHANNEL_1, DAC_ALIGN_12B_R, (uint32_t) trg_params.binary);
		HAL_DAC_Start(gpDac, DAC_CHANNEL_1);
		HAL_OPAMP_Start(gpOpamp);
		dac_params.prescaler = (uint16_t) gpTimDac->Instance->PSC;
		dac_params.counter = gpTimDac->Instance->ARR;
		dac_params.start_index = 0;
		dac_params.stop_index = DAC_MEMORY_SIZE-1;
		dac_params.repeat_times = 0;  // infinity
		dac_params.off_binary = 0;
		dac_params.state = DAC_STATE_OFF;
		disable_dac_trigger();
		HAL_DAC_Start(gpDac, DAC_CHANNEL_2);
		HAL_DAC_SetValue(gpDac, DAC_CHANNEL_2, DAC_ALIGN_12B_R, (uint32_t) dac_params.off_binary);
		adc_params.prescaler = (uint16_t) gpTimAdc->Instance->PSC;
		adc_params.counter = (uint16_t) gpTimAdc->Instance->ARR;
		adc_params.state = ADC_STATE_OFF;
		__HAL_COMP_DISABLE(gpComp);
		for(i=0; i<GPIO_PINS; ++i){
			gpio_params.state[i] = GPIO_L;
		}
		set_gpio(gpio_params.state);
		set_led(0);
	}
}

/* コマンドの実行
 *
 * コマンド構造体で指定されたコマンドを実行する。
 * コマンド引数は正当と見なして、値の正当性確認はしない。
 * あらかじめ、command.c 側で値の正当性確認しておくこと。
 *
 * [引数]
 * 		command_def : コマンド構造体ポインタ
 *
 * 	[戻値]
 * 		0 : コマンド不備で実行されなかった
 * 		1 : 正常にコマンド実行された
 */
uint8_t action_command(const CommandTypeDef *command_def){
	uint32_t ack[COMMAND_RETURNS];
	uint16_t i;

	// エラーフラグをクリア
	flg_adc_error=0;
	flg_dac_error=0;

	if(! command_def->effective){
		// ACTION_FALSE
	}
	// SYSTEM_MODE_DIRECT --------------------------------------------------------
	else if(system_mode == SYSTEM_MODE_DIRECT){
		// No ACK
		return ACTION_FALSE;
	}
	// SYSTEM_MODE_LOG ------------------------------------------------------------
	else if(system_mode == SYSTEM_MODE_LOG){
		if(command_def->command_header == COMMAND_SET_HEADER &&
				command_def->command_type == COMMAND_MODE_FREE){
			system_mode = SYSTEM_MODE_FREE;
			HAL_TIM_Base_Stop_IT(gpTimMonitor);
			rewind_tim(gpTimMonitor);
			// No ACK
			return ACTION_TRUE;
		}
		else{
			// No ACK
			return ACTION_FALSE;
		}
	}
	// SYSTEM_MODE_FREE -----------------------------------------------------------
	else if(system_mode == SYSTEM_MODE_FREE){
		// コマンド先頭文字が '-' のコマンド類処理 ++++++++++++++++++++++
		if(command_def->command_header == COMMAND_GET_HEADER){
			switch(command_def->command_type){
			// ====================
			// -l: システムモード SYSTEM_MODE_LOG におけるログデータ条件を取得
			// ====================
			case COMMAND_LOG_PARAM:		// -l
				ack[0] = (uint32_t) log_params.prescaler;
				ack[1] = (uint32_t) log_params.counter;
				uart_hex_send(ack, 2);
				return ACTION_TRUE;
				break;
			// ====================
			// -p: PWM出力の条件を取得
			// ====================
			case COMMAND_PWM_PARAM:	// -p
				ack[0] = (uint32_t) pwm_params.prescaler;
				ack[1] = pwm_params.counter;
				ack[2] = pwm_params.pulse;
				uart_hex_send(ack, 3);
				return ACTION_TRUE;
				break;
			// ====================
			// -P: PWM出力の状態を取得
			// ====================
			case COMMAND_PWM_STATE:		// -P
				ack[0] = (uint32_t) pwm_params.state;
				uart_hex_send(ack, 1);
				return ACTION_TRUE;
				break;
			// ====================
			// -t: トリガ閾値電圧を取得
			// ====================
			case COMMAND_TRG_PARAM:		// -t
				ack[0] = (uint32_t) trg_params.binary;
				uart_hex_send(ack, 1);
				return ACTION_TRUE;
				break;
			// ====================
			// -d: DAC Ch.A 出力の条件を取得
			// ====================
			case COMMAND_DAC_PARAM:		// -d
				ack[0] = (uint32_t) dac_params.prescaler;
				ack[1] = dac_params.counter;
				ack[2] = (uint32_t) dac_params.start_index;
				ack[3] = (uint32_t) dac_params.stop_index;
				ack[4] = (uint32_t) dac_params.repeat_times;
				ack[5] = (uint32_t) dac_params.off_binary;
				uart_hex_send(ack, 6);
				return ACTION_TRUE;
				break;
			// ====================
			// -e: DAC Ch.A 出力 OFF 時の DACバイナリ値のみ取得
			// ====================
			case COMMAND_DAC_OFFBINARY:	// -e
				ack[0] = (uint32_t) dac_params.off_binary;
				uart_hex_send(ack, 1);
				return ACTION_TRUE;
				break;
			// ====================
			// -D: DAC Ch.A 出力の状態を取得
			// ====================
			case COMMAND_DAC_STATE:		// -D
				ack[0] = (uint32_t) dac_params.state;
				uart_hex_send(ack, 1);
				return ACTION_TRUE;
				break;
			// ====================
			// -a: ADC Ch.A/B 電圧サンプリングの条件を取得
			// ====================
			case COMMAND_ADC_PARAM:		// -a
				ack[0] = (uint32_t) adc_params.prescaler;
				ack[1] = (uint32_t) adc_params.counter;
				uart_hex_send(ack, 2);
				return ACTION_TRUE;
				break;
			// ====================
			// -r: ADC Ch.A/B 電圧サンプリングデータ(ADC波形メモリ) を取得
			// ====================
			case COMMAND_ADC_MEMORY:		// -r
				uart_bin_send(adc_memory1, ADC_MEMORY_SIZE);
				uart_bin_send(adc_memory2, ADC_MEMORY_SIZE);
				return ACTION_TRUE;
				break;
			//  ====================
			// -A: ADC Ch.A/B 電圧サンプリングの状態を取得
			// ====================
			case COMMAND_ADC_STATE:		// -A
				ack[0] = (uint32_t) adc_params.state;
				uart_hex_send(ack, 1);
				return ACTION_TRUE;
				break;
			//  ====================
			// -g: GPIO出力の論理状態を取得
			//  ====================
			case COMMAND_GPIO_PARAM:		// -g
				for(i=0; i<GPIO_PINS; ++i){
					ack[i] = (uint32_t) gpio_params.state[i];
				}
				uart_hex_send(ack, GPIO_PINS);
				return ACTION_TRUE;
				break;
			//  ====================
			// -m: MONITOR Ch.A/B/C/D の電圧取得
			// ====================
			case COMMAND_MON_VALUE:			// -m
				flg_monitor = 0;
				HAL_ADC_Start_DMA(gpMonitor, (uint32_t *)monitor_memory, MONITOR_PINS);
				while(! flg_monitor){
					// Wait...
				}
				for(i=0; i<MONITOR_PINS; ++i){
					ack[i] = (uint32_t) monitor_memory[i];
				}
				uart_hex_send(ack, MONITOR_PINS);
				return ACTION_TRUE;
				break;
			// ====================
			// <TODO> -系のコマンドを追加する場合は、ここに追記する
			// ====================
			//
			// case XXXX:
			//     ...
			//     break;
			default:
				// ACTION_FALSE
				break;
			}
		}
		// コマンド先頭文字が '+' のコマンド類処理 ++++++++++++++++++++++
		else if(command_def->command_header == COMMAND_SET_HEADER){
			switch(command_def->command_type){
			// ====================
			// +l: Monitor電圧ロギングモード(システムモード SYSTEM_MODE_LOG) におけるログデータ条件を設定
			// ====================
			case COMMAND_LOG_PARAM:		// +l
				log_params.prescaler = (uint16_t) command_def->params[0];
				log_params.counter = (uint16_t) command_def->params[1];
				gpTimMonitor->Instance->PSC = command_def->params[0];
				gpTimMonitor->Instance->ARR = command_def->params[1];
				gpTimMonitor->Instance->CNT = command_def->params[1];
				uart_hex_send(ack, 0);
				return ACTION_TRUE;
				break;
			// ====================
			// +p: PWM出力の条件を設定
			// ====================
			case COMMAND_PWM_PARAM:	// +p
				if(pwm_params.state == PWM_STATE_ON){
					HAL_TIM_PWM_Stop(gpTimPwm, TIM_CHANNEL_3);
				}
				pwm_params.prescaler = (uint16_t) command_def->params[0];
				pwm_params.counter = command_def->params[1];
				pwm_params.pulse = command_def->params[2];
				gpTimPwm->Instance->PSC = command_def->params[0];
				gpTimPwm->Instance->ARR = command_def->params[1];
				gpTimPwm->Instance->CNT = command_def->params[1];
				gpTimPwm->Instance->CCR3 = command_def->params[2];
				if(pwm_params.state == PWM_STATE_ON){
					HAL_TIM_PWM_Start(gpTimPwm, TIM_CHANNEL_3);
				}
				uart_hex_send(ack, 0);
				return ACTION_TRUE;
				break;
			// ====================
			// +P: PWM出力の状態を設定
			// ====================
			case COMMAND_PWM_STATE:		// +P
				if(pwm_params.state != (uint8_t) command_def->params[0]){
					pwm_params.state = (uint8_t) command_def->params[0];
					if(pwm_params.state == PWM_STATE_ON){
						HAL_TIM_PWM_Start(gpTimPwm, TIM_CHANNEL_3);
					}
					else{
						HAL_TIM_PWM_Stop(gpTimPwm, TIM_CHANNEL_3);
					}
				}
				uart_hex_send(ack, 0);
				return ACTION_TRUE;
				break;
			// ====================
			// +t: トリガ閾値電圧を設定
			// ====================
			case COMMAND_TRG_PARAM:		// +t
				trg_params.binary = (uint16_t) command_def->params[0];
				HAL_DAC_SetValue(gpDac, DAC_CHANNEL_1, DAC_ALIGN_12B_R, command_def->params[0]);
				uart_hex_send(ack, 0);
				return ACTION_TRUE;
				break;
			// ====================
			// +d: DAC Ch.A 出力の条件を設定
			// ====================
			case COMMAND_DAC_PARAM:		// +d
				if(dac_params.state == DAC_STATE_ON){
					HAL_TIM_Base_Stop(gpTimDac);
					disable_dac_trigger();
				}
				dac_params.prescaler = (uint16_t) command_def->params[0];
				dac_params.counter = command_def->params[1];
				dac_params.start_index = (uint16_t) command_def->params[2];
				dac_params.stop_index = (uint16_t) command_def->params[3];
				dac_params.repeat_times = (uint16_t) command_def->params[4];
				dac_params.off_binary = (uint16_t) command_def->params[5];
				dac_params.now_times = 0;
				gpTimDac->Instance->PSC = command_def->params[0];
				gpTimDac->Instance->ARR = command_def->params[1];
				gpTimDac->Instance->CNT = command_def->params[1];
				if(dac_params.state == DAC_STATE_ON){
					enable_dac_trigger();
					HAL_DAC_Start_DMA(gpDac, DAC_CHANNEL_2, (uint32_t *) (dac_memory+dac_params.start_index), (uint32_t) (dac_params.stop_index-dac_params.start_index+1), DAC_ALIGN_12B_R);
					HAL_TIM_Base_Start(gpTimDac);
				}
				else{
					HAL_DAC_SetValue(gpDac, DAC_CHANNEL_2, DAC_ALIGN_12B_R, command_def->params[5]);
				}
				uart_hex_send(ack, 0);
				return ACTION_TRUE;
				break;
			// ====================
			// +e: DAC Ch.A 出力 OFF 時の DACバイナリ値のみ設定
			// ====================
			case COMMAND_DAC_OFFBINARY:
				dac_params.off_binary = (uint16_t) command_def->params[0];
				if(dac_params.state == DAC_STATE_OFF){
					HAL_DAC_SetValue(gpDac, DAC_CHANNEL_2, DAC_ALIGN_12B_R, command_def->params[0]);
				}
				uart_hex_send(ack, 0);
				return ACTION_TRUE;
				break;
			// ====================
			// +s: DAC Ch.A 出力のDAC波形メモリへのデータ転送開始
			// ====================
			case COMMAND_DAC_MEMORY:		// +s
				if(dac_params.state == DAC_STATE_ON){
					HAL_TIM_Base_Stop(gpTimDac);
					disable_dac_trigger();
					rewind_tim(gpTimDac);
					HAL_DAC_SetValue(gpDac, DAC_CHANNEL_2, DAC_ALIGN_12B_R, (uint32_t) dac_params.off_binary);
					dac_params.state = DAC_STATE_OFF;
					dac_params.now_times = 0;
				}
				// DAC波形メモリをゼロ初期化
				memset(dac_memory, 0, sizeof(dac_memory));
				uart_hex_send(ack, 0);
				// UART受信をバイナリ受信モードとして、直接メモリに書込む (波形メモリは1点2byte なので、指定受信バイト数は2倍)
				uart_rx_set_direct((uint8_t *) dac_memory, (uint16_t) command_def->params[0] << 1);
				return ACTION_TRUE;
				break;
			// ====================
			// +D: DAC Ch.A 出力の状態を設定
			// ====================
			case COMMAND_DAC_STATE:		// +D
				if(dac_params.state != (uint8_t) command_def->params[0]){
					dac_params.state = (uint8_t) command_def->params[0];
					if(dac_params.state == DAC_STATE_OFF){
						HAL_TIM_Base_Stop(gpTimDac);
						disable_dac_trigger();
						rewind_tim(gpTimDac);
						HAL_DAC_SetValue(gpDac, DAC_CHANNEL_2, DAC_ALIGN_12B_R, (uint32_t) dac_params.off_binary);
						dac_params.now_times = 0;
					}
					else{
						dac_params.now_times = 0;
						enable_dac_trigger();
						HAL_DAC_Start_DMA(gpDac, DAC_CHANNEL_2, (uint32_t *) (dac_memory+dac_params.start_index), (uint32_t) (dac_params.stop_index-dac_params.start_index+1), DAC_ALIGN_12B_R);
						HAL_TIM_Base_Start(gpTimDac);
					}
				}
				uart_hex_send(ack, 0);
				return ACTION_TRUE;
				break;
			// ====================
			// +a: ADC Ch.A/B 電圧サンプリングの条件を設定
			// ====================
			case COMMAND_ADC_PARAM:
				if(adc_params.state == ADC_STATE_OFF || adc_params.state == ADC_STATE_COMPLETE){
					adc_params.prescaler = (uint16_t) command_def->params[0];
					adc_params.counter = (uint16_t) command_def->params[1];
					gpTimAdc->Instance->PSC = command_def->params[0];
					gpTimAdc->Instance->ARR = command_def->params[1];
					gpTimAdc->Instance->CNT = command_def->params[1];
					uart_hex_send(ack, 0);
					return ACTION_TRUE;
				}
				else{
					// ACTION_FALSE
				}
				break;
			// ====================
			// +A: ADC Ch.A/B 電圧サンプリングの状態を設定
			// ====================
			case COMMAND_ADC_STATE:
				if(command_def->params[0] == ADC_STATE_OFF){
					HAL_COMP_Stop(gpComp);
					HAL_TIM_Base_Stop(gpTimAdc);
					rewind_tim(gpTimAdc);
					HAL_ADC_Stop_DMA(gpAdcChA);
					HAL_ADC_Stop_DMA(gpAdcChB);
					adc_params.state = ADC_STATE_OFF;
					uart_hex_send(ack, 0);
					return ACTION_TRUE;
				}
				else if(command_def->params[0] == ADC_STATE_WAIT_RISE){
					if(adc_params.state == ADC_STATE_WAIT_RISE){
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
					}
					switch(adc_params.state){
					case ADC_STATE_OFF:
					case ADC_STATE_COMPLETE:
						__HAL_COMP_COMP2_EXTI_ENABLE_RISING_EDGE();
						__HAL_COMP_COMP2_EXTI_DISABLE_FALLING_EDGE();
						HAL_COMP_Start(gpComp);
						adc_params.state = ADC_STATE_WAIT_RISE;
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
						break;
					case ADC_STATE_WAIT_FALL:
						__HAL_COMP_COMP2_EXTI_ENABLE_RISING_EDGE();
						__HAL_COMP_COMP2_EXTI_DISABLE_FALLING_EDGE();
						adc_params.state = ADC_STATE_WAIT_RISE;
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
						break;
					case ADC_STATE_SAMPLING:
						HAL_TIM_Base_Stop(gpTimAdc);
						rewind_tim(gpTimAdc);
						HAL_ADC_Stop_DMA(gpAdcChA);
						HAL_ADC_Stop_DMA(gpAdcChB);
						__HAL_COMP_COMP2_EXTI_ENABLE_RISING_EDGE();
						__HAL_COMP_COMP2_EXTI_DISABLE_FALLING_EDGE();
						HAL_COMP_Start(gpComp);
						adc_params.state = ADC_STATE_WAIT_RISE;
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
						break;
					default:
						// ACTION_FALSE
						break;
					}
				}
				else if(command_def->params[0] == ADC_STATE_WAIT_FALL){
					if(adc_params.state == ADC_STATE_WAIT_FALL){
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
					}
					switch(adc_params.state){
					case ADC_STATE_OFF:
					case ADC_STATE_COMPLETE:
						__HAL_COMP_COMP2_EXTI_DISABLE_RISING_EDGE();
						__HAL_COMP_COMP2_EXTI_ENABLE_FALLING_EDGE();
						HAL_COMP_Start(gpComp);
						adc_params.state = ADC_STATE_WAIT_FALL;
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
						break;
					case ADC_STATE_WAIT_RISE:
						__HAL_COMP_COMP2_EXTI_DISABLE_RISING_EDGE();
						__HAL_COMP_COMP2_EXTI_ENABLE_FALLING_EDGE();
						adc_params.state = ADC_STATE_WAIT_FALL;
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
						break;
					case ADC_STATE_SAMPLING:
						HAL_TIM_Base_Stop(gpTimAdc);
						rewind_tim(gpTimAdc);
						HAL_ADC_Stop_DMA(gpAdcChA);
						HAL_ADC_Stop_DMA(gpAdcChB);
						__HAL_COMP_COMP2_EXTI_DISABLE_RISING_EDGE();
						__HAL_COMP_COMP2_EXTI_ENABLE_FALLING_EDGE();
						HAL_COMP_Start(gpComp);
						adc_params.state = ADC_STATE_WAIT_FALL;
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
						break;
					default:
						// ACTION_FALSE
						break;
					}
				}
				else if(command_def->params[0] == ADC_STATE_SAMPLING){
					if(adc_params.state == ADC_STATE_SAMPLING){
						uart_hex_send(ack, 0);
						return ACTION_TRUE;
					}
					switch(adc_params.state){
						case ADC_STATE_OFF:
						case ADC_STATE_COMPLETE:
							HAL_TIM_Base_Stop(gpTimAdc);
							rewind_tim(gpTimAdc);
							HAL_ADC_Start_DMA(gpAdcChA, (uint32_t *) adc_memory1, ADC_MEMORY_SIZE);
							HAL_ADC_Start_DMA(gpAdcChB, (uint32_t *) adc_memory2, ADC_MEMORY_SIZE);
							HAL_TIM_Base_Start(gpTimAdc);
							adc_params.state = ADC_STATE_SAMPLING;
							uart_hex_send(ack, 0);
							return ACTION_TRUE;
							break;
						case ADC_STATE_WAIT_RISE:
						case ADC_STATE_WAIT_FALL:
							HAL_COMP_Stop(gpComp);
							HAL_TIM_Base_Stop(gpTimAdc);
							rewind_tim(gpTimAdc);
							HAL_ADC_Start_DMA(gpAdcChA, (uint32_t *) adc_memory1, ADC_MEMORY_SIZE);
							HAL_ADC_Start_DMA(gpAdcChB, (uint32_t *) adc_memory2, ADC_MEMORY_SIZE);
							HAL_TIM_Base_Start(gpTimAdc);
							adc_params.state = ADC_STATE_SAMPLING;
							uart_hex_send(ack, 0);
							return ACTION_TRUE;
							break;
						default:
							// ACTION_FALSE
							break;
						}
				}
				else{
					// ACTION_FALSE
				}
				break;
			// ====================
			// +g: GPIO出力の論理状態を設定、即座に出力値が切り替わる
			// ====================
			case COMMAND_GPIO_PARAM:
				for(i=0; i<GPIO_PINS; ++i){
					if(command_def->params[i] < GPIO_X){
						gpio_params.state[i] = (uint8_t)  command_def->params[i];
					}
				}
				set_gpio(gpio_params.state);
				uart_hex_send(ack, 0);
				return ACTION_TRUE;
				break;
			// ====================
			// +F: システムモードを SYSTEM_MODE_FREE に移行させる
			// ====================
			case COMMAND_MODE_FREE:
				// No ACK
				return ACTION_TRUE;
				break;
			// ====================
			// +L: システムモードを SYSTEM_MODE_LOG に移行させる
			// ====================
			case COMMAND_MODE_LOG:
				flg_monitor = 0;
				system_mode = SYSTEM_MODE_LOG;
				HAL_TIM_Base_Start_IT(gpTimMonitor);
				// SYTEM_MODE_LOG の挙動
				//
				// 1. TIM4 を割込み有りで、アクティブにする
				// 2. イベント UPDATE_EVENT で周期的に割込みがかかる
				// 3. 割込みイベント処理関数 HAL_TIM_PeriodElapsedCallback (action.cで定義) が呼ばれる。
				//    この処理内で、AD変換予約をする。DMAを使った Scan conversion で、4ch 分のAD変換を予約する。
				// 4. DMA で 4ch の変換が終わったら、割込みイベント処理関数 HAL_ADC_ConvCpltCallback (action.cで定義) が呼ばれる。
				//    この処理内で、フラグ flg_monitor をインクリメントする。
				// 5. main ループ内で action_polling 関数を呼び出してもらい、ポーリング処理。同関数内でフラグ flg_monitor を
				//    チェックして、非0なら DMA で転送された AD変換4ch 分の値を、UART 出力する。再度 flg_monitor を 0 にする。
				// 6. 再び 2. に戻る。

				// No ACK
				return ACTION_TRUE;
				break;
				// ====================
				// <TODO> +系のコマンドを追加する場合は、ここに追記する
				// ====================
				//
				// case XXXX:
				//     ...
				//     break;
			default:
				// ACTION_FALSE
				break;
			}
		}
		else{
			// ACTION_FALSE
		}
	}
	else{
		// ACTION_FALSE
	}

	// ACTION_FALSE 時の処理
	HAL_UART_Transmit(gpUart, (uint8_t *)"N\r", 2, 100);
	return ACTION_FALSE;
}

/* 待ち時間を伴うコマンド実行のポーリング処理
 *
 * SYSTEM_MODE_LOG 時のADC結果出力、およびエラー表示
 */
void action_polling(void){
	uint8_t i;
	uint16_t ack[MONITOR_PINS];

	if(system_mode == SYSTEM_MODE_LOG && flg_monitor > 0){
		for(i=0; i<MONITOR_PINS; ++i){
			ack[i] = monitor_memory[i];
		}
		uart_dec_send(ack, MONITOR_PINS);
		if(flg_monitor > 1){
			// Detect Monitor Overwrite Error, Do nothing
		}
		flg_monitor = 0;
	}
	if(flg_adc_error > 0){
		// Detect ADC Error, Do nothing
		flg_adc_error = 0;
	}
	if(flg_dac_error > 0){
		// Detect DAC Error, Do nothing
		flg_dac_error = 0;
	}
}

/* (内部関数) GPIO 4ch の設定
 *
 * [引数]
 * 		chA, chB, chC, chD の順に uint8_t 配列
 * 		0: Low, 1: High, 2<: 変化なし
 */
static void set_gpio(const uint8_t *parameter){
	if(parameter[0] < GPIO_X){
		HAL_GPIO_WritePin(GPIOA, GPIO_PIN_8, parameter[0]);	// Ch.A
	}
	if(parameter[1] < GPIO_X){
		HAL_GPIO_WritePin(GPIOA, GPIO_PIN_9, parameter[1]);	// Ch.B
	}
	if(parameter[2] < GPIO_X){
		HAL_GPIO_WritePin(GPIOC, GPIO_PIN_7, parameter[2]);	// Ch.C
	}
	if(parameter[3] < GPIO_X){
		HAL_GPIO_WritePin(GPIOB, GPIO_PIN_6, parameter[3]);	// Ch.D
	}
}

/* (内部関数) Green LEDの設定
 *
 * [引数]
 * 		0: Low, 1: High, 2<: 変化なし
 */
static void set_led(uint8_t level){
	if(level < GPIO_X){
		HAL_GPIO_WritePin(GPIOA, GPIO_PIN_5, level);	// PA5
	}
}

/* (内部関数) UART の8桁Hex文字列形式の応答送信
 *
 * 先頭文字 'Y' を付加し、引数を8桁Hex文字列で送信する。終端文字は COMMAND_TERMINAL
 *
 * [引数]
 * 		parameter : 戻値の変数配列
 * 		size : 有効な変数の数 (10以下)
 */
static void uart_hex_send(const uint32_t *parameter, uint8_t size){
	uint8_t send_str[COMMAND_RETURNS*8+2];
	uint8_t str_index, parameter_index, i, hex;

	if(size > COMMAND_RETURNS){
		// Do nothing
	}

	str_index = 0;
	send_str[str_index++] = 'Y';
	for(parameter_index = 0; parameter_index < size; ++parameter_index){
		for(i=0; i<8; ++i){
			hex = (uint8_t) ((parameter[parameter_index] >> (28-4*i)) & 0x000ful);
			if(hex < 10){
				send_str[str_index++] = hex + '0';
			}
			else{
				send_str[str_index++] = hex - 10 + 'A';
			}
		}
	}
	send_str[str_index++] = COMMAND_TERMINAL;

	// timeout 100ms
	HAL_UART_Transmit(gpUart, send_str, str_index, 100);
}

/* (内部関数) UART の10進数文字列形式の応答送信
 *
 * 引数をカンマ区切りの10進数文字列で送信する。終端文字は CR+LF ("\r\n")
 *
 * [引数]
 * 		parameter : 戻値の変数配列
 * 		size : 有効な変数の数 (10以下)
 */
static void uart_dec_send(const uint16_t *parameter, uint8_t size){
	uint8_t send_str[COMMAND_RETURNS*6+2];
	uint8_t str_index, parameter_index, effective;
	uint16_t range, dec, tmp_parameter;

	if(size > COMMAND_RETURNS || size == 0){
		// Do nothing
	}

	str_index = 0;
	for(parameter_index = 0; parameter_index < size; ++parameter_index){
		tmp_parameter = parameter[parameter_index];
		effective = 0;
		for(range=10000; range>=1; range/=10){
			dec = tmp_parameter/range;
			tmp_parameter %= range;
			if(dec == 0 && effective ==0 && range > 1){
				// Do nothing
			}
			else{
				effective = 1;
				send_str[str_index++] = dec + '0';
			}
		}
		send_str[str_index++] = ',';
	}
	--str_index;
	send_str[str_index++] = '\r';
	send_str[str_index++] = '\n';

	// timeout 100ms
	HAL_UART_Transmit(gpUart, send_str, str_index, 100);
}

/* (内部関数) UART のバイナリ応答送信
 *
 * 引数の無符号16bitをリトルエンディアン形式で送信。終端文字は無し。
 *
 * [引数]
 * 		parameter : 戻値の変数配列
 * 		size : 有効な変数の数 (32768未満)
 */
static void uart_bin_send(const uint16_t *parameter, uint16_t size){
	if(size < 32768){
		HAL_UART_Transmit(gpUart, (uint8_t*) parameter, size<<1, 10000);
	}
}

/* (内部関数) DAC Ch.A のトリガ無効化 */
static void disable_dac_trigger(void){
	gpDac->Instance->CR &= ~(DAC_CR_DMAEN1 << (DAC_CHANNEL_2 & 0x10UL));
	HAL_DMA_Abort(gpDac->DMA_Handle2);
	/* Disable the DAC DMA underrun interrupt */
	__HAL_DAC_DISABLE_IT(gpDac, DAC_IT_DMAUDR2);
	// トリガ無効化
	gpDac->Instance->CR &= 0xfffbfffful;
}

/* (内部関数) DAC Ch.A のトリガ有効化 */
static void enable_dac_trigger(void){
	gpDac->Instance->CR |= 0x00040000ul;
}

/* (内部関数) TIM のカウンタ巻き */
static void rewind_tim(TIM_HandleTypeDef *pTim){
	pTim->Instance->CNT = pTim->Instance->ARR;
}

void HAL_ADC_ConvCpltCallback(ADC_HandleTypeDef *hadc){
	if(hadc == gpMonitor){
		++flg_monitor;
	}
	else if(hadc == gpAdcChA){
		HAL_ADC_Stop_DMA(gpAdcChA);
		if(adc_params.state == ADC_STATE_COMPLETE){
			HAL_TIM_Base_Stop(gpTimAdc);
			rewind_tim(gpTimAdc);
		}
		adc_params.state = ADC_STATE_COMPLETE;
	}
	else if(hadc == gpAdcChB){
		HAL_ADC_Stop_DMA(gpAdcChB);
		if(adc_params.state == ADC_STATE_COMPLETE){
			HAL_TIM_Base_Stop(gpTimAdc);
			rewind_tim(gpTimAdc);
		}
		adc_params.state = ADC_STATE_COMPLETE;
	}
}

void HAL_ADC_ErrorCallback(ADC_HandleTypeDef *hadc){
	++flg_adc_error;
}

void HAL_DACEx_ConvCpltCallbackCh2(DAC_HandleTypeDef *hdac){
	if(dac_params.repeat_times != 0){
		++dac_params.now_times;
		if(dac_params.now_times == dac_params.repeat_times){
			HAL_TIM_Base_Stop(gpTimDac);
			disable_dac_trigger();
			rewind_tim(gpTimDac);
			HAL_DAC_SetValue(gpDac, DAC_CHANNEL_2, DAC_ALIGN_12B_R, (uint32_t) dac_params.off_binary);
			dac_params.state = DAC_STATE_OFF;
			dac_params.now_times = 0;
		}
	}
}

void HAL_DACEx_ErrorCallbackCh2(DAC_HandleTypeDef *hdac){
	++flg_dac_error;
}

void HAL_DACEx_DMAUnderrunCallbackCh2(DAC_HandleTypeDef *hdac){
	++flg_dac_error;
}

void HAL_COMP_TriggerCallback(COMP_HandleTypeDef *hcomp){
	if(adc_params.state == ADC_STATE_WAIT_RISE && HAL_COMP_GetOutputLevel(gpComp) == COMP_OUTPUT_LEVEL_HIGH){
		HAL_ADC_Start_DMA(gpAdcChA, (uint32_t *) adc_memory1, ADC_MEMORY_SIZE);
		HAL_ADC_Start_DMA(gpAdcChB, (uint32_t *) adc_memory2, ADC_MEMORY_SIZE);
		HAL_TIM_Base_Start(gpTimAdc);
		HAL_COMP_Stop(gpComp);
		adc_params.state = ADC_STATE_SAMPLING;
	}
	else if(adc_params.state == ADC_STATE_WAIT_FALL && HAL_COMP_GetOutputLevel(gpComp) == COMP_OUTPUT_LEVEL_LOW){
		HAL_ADC_Start_DMA(gpAdcChA, (uint32_t *) adc_memory1, ADC_MEMORY_SIZE);
		HAL_ADC_Start_DMA(gpAdcChB, (uint32_t *) adc_memory2, ADC_MEMORY_SIZE);
		HAL_TIM_Base_Start(gpTimAdc);
		adc_params.state = ADC_STATE_SAMPLING;
	}
}

void HAL_TIM_PeriodElapsedCallback(TIM_HandleTypeDef *htim){
	if(htim == gpTimMonitor){
		HAL_ADC_Start_DMA(gpMonitor, (uint32_t *) monitor_memory, MONITOR_PINS);
	}
}
