
#include "command.h"

static uint8_t hexstr2val(const uint8_t *command_str, uint8_t param_index, uint32_t *value);


/* コマンド文字列から、コマンド構造体(CommandTypeDef) への変換
 *
 * [コマンド形式]
 * 		1. コマンドの先頭文字は ‘+’ か ‘-’ で、この文字以降をコマンドと見なす。大別して + は設定系、- は取得系コマンドを意味する。
 *		2. その後の、英字1文字はコマンド種別文字とする。
 *		3. その後に、0個以上の複数個のコマンド引数を連ねる。1引数当たり、0埋め8文字固定のHEX文字列とする。
 *		4. コマンドの終端文字は CR ('\r')
 *
 * 		1つのコマンド引数は、8桁のHex文字列(32bit) として、複数のパラメータを連結する。文字種は、0-9, A-F, a-f を許容する。
 * 		パラメータ個数は、コマンド種別により固定とする。過不足あれば、不当なコマンドとみなす。
 * 		<コマンド例>  "+p00000003000001AD0000001F\r" => 設定系のコマンドで、コマンド種別 p , 引数1=0x03, 引数2=0x1AD, 引数3=0x1F
 *
 * 	[引数]
 * 		command_str : コマンド文字列先頭アドレス
 * 		command_size : コマンド文字列有効文字数(終端文字 CRを含む)
 * 		command_def : コマンド文字列から変換した結果を格納する、コマンド構造体のアドレス
 *
 * 	[戻値]
 * 		0 : 不当なコマンド文字列
 * 		1 : 正常なコマンド文字列で、有効なコマンド構造体に変換されている
 */
uint8_t command_analysis(const uint8_t *command_str, uint8_t command_size, CommandTypeDef *command_def){
	uint8_t num_params, i;

	command_def->effective = COMMAND_FALSE;
	if(command_size < 3 || (command_size-3) % 8 != 0){
		return COMMAND_FALSE;
	}

	command_def->command_header = command_str[0];
	command_def->command_type = command_str[1];
	num_params = (command_size-3) / 8;
	if(num_params > COMMAND_PARAMETERS){
		return COMMAND_FALSE;
	}

	// パラメータ変換
	for(i = 0; i < num_params; ++i){
		if(hexstr2val(command_str, i, command_def->params + i) == COMMAND_FALSE){
			return COMMAND_FALSE;
		}
	}

	// コマンド種別、コマンド引数の正当性チェック
	if(command_def->command_header == COMMAND_GET_HEADER){
		// <TODO> -系のコマンドを追加する場合は、ここに追記する。
		//                 コマンド引数ある場合は、個数と値範囲チェックする。
		if(num_params == 0){
			switch(command_def->command_type){
			case COMMAND_LOG_PARAM:
			case COMMAND_PWM_PARAM:
			case COMMAND_PWM_STATE:
			case COMMAND_TRG_PARAM:
			case COMMAND_DAC_PARAM:
			case COMMAND_DAC_OFFBINARY:
			case COMMAND_DAC_STATE:
			case COMMAND_ADC_PARAM:
			case COMMAND_ADC_MEMORY:
			case COMMAND_ADC_STATE:
			case COMMAND_GPIO_PARAM:
			case COMMAND_MON_VALUE:
				command_def->effective = COMMAND_TRUE;
				break;
			default:
				break;
			}
		}
	}
	else if(command_def->command_header == COMMAND_SET_HEADER){
		// <TODO> +系のコマンドを追加する場合は、ここに追記する。
		//                コマンド引数ある場合は、個数と値範囲チェックする。
		switch(command_def->command_type){
			case COMMAND_MODE_FREE:
				if(num_params == 0){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_MODE_LOG:
				if(num_params == 0){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_LOG_PARAM:
				// Prescaler(16bit), Counter(16bit)
				if(num_params == 2 &&
						command_def->params[0] <= 0xffff && command_def->params[1] <= 0xffff &&
						(command_def->params[0]+1) * (command_def->params[1]+1) >= 800000ul){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_PWM_PARAM:
				// Prescaler(16bit), Counter(32bit), Pulse(32bit)
				if(num_params == 3 && command_def->params[0] <= 0xffff){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_PWM_STATE:
				if(num_params == 1 &&
						(command_def->params[0] == PWM_STATE_OFF || command_def->params[0] == PWM_STATE_ON)){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_TRG_PARAM:
				if(num_params == 1 && command_def->params[0] <= 0x0fff){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_DAC_PARAM:
				// Prescaler(16bit), Counter(32bit), Start index, Stop index, Repeat times(16bit), OFF value(12bit)
				if(num_params == 6 &&
						command_def->params[0] <= 0xffff &&
						(command_def->params[0]+1)*(command_def->params[1]+1) >= 40 &&
						command_def->params[2] < DAC_MEMORY_SIZE &&
						command_def->params[3] < DAC_MEMORY_SIZE &&
						command_def->params[2] < command_def->params[3] &&
						command_def->params[4] < 0xffff &&
						command_def->params[5] <= 0x0fff){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_DAC_OFFBINARY:
				// OFF value(12bit)
				if(num_params == 1 && command_def->params[0] <= 0x0fff){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_DAC_MEMORY:
				// Size
				if(num_params == 1 && command_def->params[0] <= DAC_MEMORY_SIZE){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_DAC_STATE:
				if(num_params == 1 &&
						(command_def->params[0] == DAC_STATE_OFF ||
						  command_def->params[0] == DAC_STATE_ON)){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_ADC_PARAM:
				// Prescaler(16bit), Counter(16bit)
				if(num_params == 2 &&
						command_def->params[0] <= 0xffff &&
						command_def->params[1] <= 0xffff &&
						(command_def->params[0]+1) * (command_def->params[1]+1) >= 20){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_ADC_STATE:
				// Start mode
				if(num_params == 1 &&
						(command_def->params[0] == ADC_STATE_OFF ||
						 command_def->params[0] == ADC_STATE_WAIT_RISE ||
						 command_def->params[0] == ADC_STATE_WAIT_FALL ||
						 command_def->params[0] == ADC_STATE_SAMPLING)){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			case COMMAND_GPIO_PARAM:
				if(num_params == 4 &&
						command_def->params[0] <= GPIO_X &&
						command_def->params[1] <= GPIO_X &&
						command_def->params[2] <= GPIO_X &&
						command_def->params[3] <= GPIO_X){
					command_def->effective = COMMAND_TRUE;
				}
				break;
			default:
				break;
		}
	} // else if(command_def->command_header == COMMAND_SET_HEADER)
	else{
		// Do nothing
	} // else

	return command_def->effective;
}


/* (内部関数) コマンドの8桁Hex文字のパラメータを数値化
 *
 * [引数]
 * 		command_str : コマンド文字列
 * 		param_index : パラメータ引数の何番目を変換するか指定 (0-)
 * 		value : コマンド文字列から変換した結果を格納する、変数アドレス
 *
 * 	[戻値]
 * 		0 : 数値表現異常により、変換されなかった
 * 		1 : 正常に変換された
 */
static uint8_t hexstr2val(const uint8_t *command_str, uint8_t param_index, uint32_t *value){
	uint8_t i, start_index, word;

	start_index = 2+param_index*8;
	*value = 0;
	for(i = 0; i < 8; ++i){
		*value <<= 4;
		word = *(command_str+start_index+i);
		if(word >= '0' && word <= '9'){
			*value += word - '0';
		}
		else if(word >= 'a' && word <= 'f'){
			*value += word - 'a' + 10;
		}
		else if(word >= 'A' && word <= 'F'){
			*value += word - 'A' + 10;
		}
		else{
			return COMMAND_FALSE;
		}
	}
	return COMMAND_TRUE;
}
