
#include "uart_rx.h"

static volatile uint8_t rx_state = UART_RX_STATE_INVALID;
static volatile uint8_t rx_buffer[UART_RX_BUFFER_SIZE];
static volatile uint8_t rx_temp;
static volatile uint16_t buffer_index;
static volatile uint8_t *direct_buffer;
static volatile uint16_t direct_index;
static volatile uint16_t direct_limit;
static UART_HandleTypeDef *rx_huart = 0;
static uint8_t rx_lock_state;

/* UART受信バックグラウンド処理初期化
 *
 * PCからのコマンドのUART受信は、バックグラウンド処理で受け付ける。
 * コマンドは、ヘッダ文字('+' or '-') から、終端文字('\r') までをスタックする。
 * ポーリング処理で、uart_rx_get_command() 関数により、確定したコマンド文字列を取得する。
 * 又は、uart_rx_get_state() 関数により、現在の受信状態を知る。
 */
void uart_rx_init(UART_HandleTypeDef *huart){
	if(! rx_huart){
		rx_huart = huart;
		buffer_index = 0;
		rx_state = UART_RX_STATE_EMPTY;
		HAL_UART_Receive_IT(rx_huart, &rx_temp, 1);
	}
}

/* UART受信状態
 *
 * [戻値]
 * 		UART_RX_STATE_INVALID			(0) : バックグラウンド処理未初期化
 * 		UART_RX_STATE_EMPTY			(1) : コマンド文字のヘッダ('+' or '-') が未検出
 * 		UART_RX_STATE_RECEIVING		(2) : コマンド文字のヘッダを受領済で、その後のコマンド文字列を受領している途中
 * 		UART_RX_STATE_COMPLETE	(3) : コマンド文字の終端文字を受領し、コマンド確定状態で待ち
 * 		UART_RX_STATE_OVERFLOW	(4) : コマンド文字列のバッファを超過している (以降の受信文字は無視、但しコマンド終端文字を受けると、UART_RX_STATE_EMPTY に戻される)
 * 		UART_RX_STATE_DIRECT			(5) : このステート中は、別途指定のメモリに直接受信バイナリを書込む、別途指定のバイナリバイト数を書込んだら、UART_RX_STATE_EMPTY に戻される
 * 		UART_RX_STATE_LOCK		  		(6) : このステート中は、受信文字を無視
 */
uint8_t uart_rx_get_state(void){
	return rx_state;
}

/* UART受信コマンドを取得
 *
 * 現在のUART受信状態が UART_RX_STATE_COMPLETE の場合、スタックされたコマンド文字列を取得する。
 *
 * [引数]
 * 		command_str : 受信コマンドを格納する配列ポインタ
 * 		max_size : 同配列ポインタの最大要素数、最大でもこの文字数しかコマンド文字列を取得できない、UART_RX_BUFFER_SIZE が理想
 *
 * 	[戻値]
 * 		command_str にコピーされた、コマンド文字列長
 * 		未だUART受信状態が UART_RX_STATE_COMPLETE 状態で無い場合は、0 になる。
 */
uint8_t uart_rx_get_command(uint8_t *command_str, uint8_t max_size){
	uint8_t copy_size, i;

	if(rx_state == UART_RX_STATE_COMPLETE){
		copy_size = (buffer_index > max_size) ? max_size : buffer_index;
		for(i = 0; i < copy_size; ++i){
			command_str[i] = rx_buffer[i];
		}
		return copy_size;
	}
	else{
		return 0;
	}
}

/* 強制的に、UART受信状態を UART_RX_STATE_EMPTY 状態に遷移させる。
 * 但し、未初期化(UART_RX_STATE_INVALID) では無効とする。*/
void uart_rx_reboot(void){
	if(rx_state != UART_RX_STATE_INVALID){
		rx_state = UART_RX_STATE_LOCK;
		buffer_index = 0;
		rx_state = UART_RX_STATE_EMPTY;
	}
}

/* 強制的に、UART受信状態を UART_RX_STATE_EMPTY 状態に遷移させる。
 * 但し、未初期化(UART_RX_STATE_INVALID) と 直接メモリ転送(UART_RX_STATE_DIRECT) 状態では無効とする。 */
void uart_rx_set_empty(void){
	if(rx_state != UART_RX_STATE_INVALID && rx_state != UART_RX_STATE_DIRECT){
		rx_state = UART_RX_STATE_LOCK;
		buffer_index = 0;
		rx_state = UART_RX_STATE_EMPTY;
	}
}

/* 強制的に、UART受信状態を UART_RX_STATE_LOCK 状態に遷移させる。元の状態は記憶しておき、uart_rx_set_unlock() で復帰可能。
 *  但し、未初期化(UART_RX_STATE_INVALID) と 直接メモリ転送(UART_RX_STATE_DIRECT) 状態では無効とする。 */
void uart_rx_set_lock(void){
	if(rx_state != UART_RX_STATE_INVALID && rx_state != UART_RX_STATE_DIRECT){
		rx_lock_state = rx_state;
		rx_state = UART_RX_STATE_LOCK;
	}
}

/* UART受信状態 UART_RX_STATE_LOCK を、ロック前の状態に戻す */
void uart_rx_set_unlock(void){
	if(rx_state != UART_RX_STATE_LOCK){
		rx_state = rx_lock_state;
	}
}

/* 強制的に、UART受信状態を UART_RX_STATE_DIRECT 状態に遷移させる
 *
 * 以降のUART 受信はバイナリモードで、指定メモリに指定長さを直接書込み、書込み終了後に、UART_RX_STATE_EMPTY 状態に遷移する
 *
 * [引数]
 * 		direct_ptr : 指定メモリの先頭アドレス
 * 		direct_size : 連続して書込む指定長さ
 */
void uart_rx_set_direct(uint8_t *direct_ptr, uint16_t direct_size){
	if(rx_state != UART_RX_STATE_INVALID){
		rx_state = UART_RX_STATE_LOCK;
		direct_buffer = direct_ptr;
		direct_limit = direct_size;
		direct_index = 0;
		rx_state = UART_RX_STATE_DIRECT;
	}
}


void HAL_UART_RxCpltCallback(UART_HandleTypeDef *huart){
	switch(rx_state){
		case UART_RX_STATE_EMPTY:
			if(rx_temp == COMMAND_SET_HEADER || rx_temp == COMMAND_GET_HEADER){
				rx_buffer[buffer_index++] = rx_temp;
				rx_state = UART_RX_STATE_RECEIVING;
			}
			break;
		case UART_RX_STATE_RECEIVING:
			rx_buffer[buffer_index++] = rx_temp;
			if(rx_temp == COMMAND_TERMINAL){
				rx_state = UART_RX_STATE_COMPLETE;
			}
			else if(buffer_index >= UART_RX_BUFFER_SIZE){
				rx_state = UART_RX_STATE_OVERFLOW;
			}
			break;
		case UART_RX_STATE_OVERFLOW:
			if(rx_temp ==  COMMAND_TERMINAL){
				// Clear buffer
				buffer_index = 0;
				rx_state = UART_RX_STATE_EMPTY;
			}
			break;
		case UART_RX_STATE_DIRECT:
			direct_buffer[direct_index++] = rx_temp;
			if(direct_index >= direct_limit){
				buffer_index = 0;
				rx_state = UART_RX_STATE_EMPTY;
			}
			break;
		default:
			// UART_RX_STATE_INVALID
			// UART_RX_STATE_COMPLETE
			// UART_RX_STATE_LOCK
			break;
	}
	// Re-entry IT
	HAL_UART_Receive_IT(rx_huart, &rx_temp, 1);
}
