from DiscoveryBase import DiscoveryBase
import utility
import math
import time

""" トランジスタ技術 2023年9月号
超安お手軽！ Nucleoのアナログ機能で作る多機能計測ステーション
様々な計測事例サンプル集

<動作確認環境>
    windows 11
    python 3.8.5

<必要ライブラリ>
    付属の DiscoveryBase.py と utility.py を同じフォルダ内に保存しておく
"""

# 全ての用例に共通する事前準備
discovery = DiscoveryBase()
discovery.open()  # 通信接続 (最後には close() メソッドで通信切断する)


# 用例1: 任意信号発生器 DAC Ch.A からの正弦波出力 ===============================
# 生成したい正弦波の希望周波数 12030 [Hz]
desired_freq = 12030
# 希望周波数に近い波形を設定する。
#   第1引数: 正弦波の希望周波数 [Hz]
#   第2引数: 1周期の波形点数
#   第3引数: バイアス電圧 [V]
#   第4引数: 振幅 [V]
#   第5引数: 位相 [degree]
#   第6引数: 繰返し回数 (0は無限繰返し)
#       時刻t の 波形(t) [V] = バイアス電圧 + 振幅 * sin(2*PI*希望周波数*t + 位相*2*PI/180)
#   ----------
#   戻値: 実際の正弦波の設定周波数 [Hz]
sine_freq = discovery.set_dac_sinwave(desired_freq, 64, 1.2, 0.73, 180, 0)
print('Ready to output {} [Hz] sine wave.\n'.format(sine_freq))
# Nucleo Discovery から任意信号波形出力 ON
discovery.enable_dac()
time.sleep(5)  # 5sec間出力
# 任意信号波形出力 OFF
discovery.disable_dac()


# 用例2: 任意信号発生器 DAC Ch.A から任意波形出力 ===============================
# 100点の波形リストを用意する。
# 本例では 1.4 - 1.8 [V] の鋸波
wave_list = [1.4+i/100*0.4 for i in range(100)]
# リストの波形データを DAC波形メモリに転送
discovery.set_dac_memory(wave_list)
# DAC出力の条件設定
#   第1引数: DAC波形メモリの更新間隔時間 [sec]
#   第2引数: DAC波形メモリの、波形始点インデックス
#   第3引数: DAC波形メモリの、波形終点インデックス
#   第4引数: 波形メモリの繰返し回数 (0 は無限繰返し)
#   第5引数: DAC出力OFF時のDAC出力電圧 0-3.3[V]
#   ----------
#   戻値: 実際の更新間隔時間 [sec]
setting_time = discovery.set_dac(0.01e-3, 0, 100-1, 200, 1.4)
# Nucleo Discovery から波形出力 ON
discovery.enable_dac()
time.sleep(1)
# 繰返し回数を経過したら、自動的に波形出力 OFF になる。


# 用例3: 任意信号発生器 DAC Ch.A から直流電圧出力 ===============================
# OFF 時の静止直流電圧を利用する。
discovery.disable_dac()
discovery.set_dac_off_voltage(1.3)  # 1.3V即時反映
time.sleep(1)
discovery.set_dac_off_voltage(2.2)  # 2.2V即時反映


# 用例4: オシロスコープ ADC Ch.A/B 電圧サンプリング ===============================
# サンプリング希望周波数設定 23k [sps]
setting_freq = discovery.set_adc_sampling_freq(23000)
print('Ready to ADC: sampling frequency {} [Hz]\n'.format(setting_freq))
# 直ぐにサンプリング開始
discovery.start_adc(DiscoveryBase.ADC_TRIGGER_NONE)
# サンプリング終了確認
while discovery.get_adc_state() != DiscoveryBase.ADC_STATE_COMPLETE:
    time.sleep(0.1)
# データ取得 各チャンネル 8192要素の電圧データ
chAdata, chBdata = discovery.get_adc_memory()
# データ可視化例
# import numpy as np
# import matplotlib.pyplot as plt
# plt.plot(np.array(chAdata))
# plt.plot(np.array(chBdata))
# plt.show()

# トリガ電圧を設定
discovery.set_trigger(1.6)  # 1.6V
# トリガ入力にトリガ信号を入力。トリガ電圧に対して、立上りエッジをトリガとしてオシロスコープ開始する。
# 但しエッジ検出処理に 22usec 必要になるため、エッジ点から 22usec 後からの記録開始になる。
discovery.start_adc(DiscoveryBase.ADC_TRIGGER_RISE)
while discovery.get_adc_state() != DiscoveryBase.ADC_STATE_COMPLETE:
    time.sleep(0.1)
# データ取得 各チャンネル 8192要素の電圧データ
chAdata, chBdata = discovery.get_adc_memory()


# 用例5: オシロスコープ ADC Ch.A FFTアナライザ ===============================
# サンプリング希望周波数設定 1M [sps] (実際に設定された周波数は、戻値 setting_freq)
desired_freq = 1e6
setting_freq = discovery.set_adc_sampling_freq(desired_freq)
# 直ぐにサンプリング開始
discovery.start_adc(DiscoveryBase.ADC_TRIGGER_NONE)
while discovery.get_adc_state() != DiscoveryBase.ADC_STATE_COMPLETE:
    time.sleep(0.1)
# データ取得
chAdata, chBdata = discovery.get_adc_memory()
import numpy as np
# ハニング窓をかけて FFT
fft_rms, fft_phase = utility.analyze_fft(chAdata, np.hanning)
# 周波数軸計算
fft_freq = utility.fft_freq(setting_freq)
# データ可視化例
# import matplotlib.pyplot as plt
# plt.plot( np.array(fft_freq[: DiscoveryBase.WAVE_MEMORY_SIZE//2]), 
#           np.array(fft_rms[: DiscoveryBase.WAVE_MEMORY_SIZE//2]) )


# 用例6: ネットワークアナライザ  =============================================
#       計測対象物の入力に ADC Ch.A / 出力に ADC Ch.B を接続
#       入力には、任意信号発生器から 100 [Hz] (1周期64点) 正弦波も入力する。
#       1.65V中心、振幅1.4V、位相0度、無限繰返し
sin_freq = discovery.set_dac_sinwave(100, 64, 1.65, 1.4, 0, 0)
discovery.enable_dac()
# 系の過渡応答待ち
time.sleep(0.1)
# サンプリング希望周波数設定 4k [sps] (実際に設定された周波数は、戻値 adc_freq)
adc_freq = discovery.set_adc_sampling_freq(4e3)
# 直ぐにサンプリング開始
discovery.start_adc(DiscoveryBase.ADC_TRIGGER_NONE)
while discovery.get_adc_state() != DiscoveryBase.ADC_STATE_COMPLETE:
    time.sleep(0.1)
discovery.disable_dac()
# データ取得
chAdata, chBdata = discovery.get_adc_memory()
# フラットトップ窓をかけて FFT (正確なピュアキャリア振幅を求めるには、フラットトップ窓が最適)
from scipy import signal
fft_rmsA, fft_phaseA = utility.analyze_fft(chAdata, signal.windows.flattop)
fft_rmsB, fft_phaseB = utility.analyze_fft(chBdata, signal.windows.flattop)
# 直流付近を除き、振幅ピークがキャリアとみなして、利得、位相差
import numpy as np
max_index = np.argmax(np.array(fft_rmsA[10: DiscoveryBase.WAVE_MEMORY_SIZE//2])) + 10
gain = fft_rmsB[max_index] / fft_rmsA[max_index]
phase = fft_phaseB[max_index] - fft_phaseA[max_index]
print('Gain: {} [dB], Phase: {} [degree], in freq: {} [Hz]'.format(20*math.log10(gain), phase, sin_freq))


# 用例7: PWM波形出力 =======================================================
# 周期 1[kHz], Duty 0.134 の PWM 波形を設定
# 実際に設定された周波数と、Duty比は戻値を参照する
set_freq, set_duty = discovery.set_pwm(1e3, 0.134)
# 設定後にイネーブルして PWM出力 ON
discovery.enable_pwm()
time.sleep(1)
# PWM出力 OFF
discovery.disable_pwm()


# 用例8: GPIO出力 ==========================================================
# Ch.A/B/C/D
discovery.set_gpio(1, 0, 0, 1)  # 即時反映

# 最後に通信切断
discovery.close()
