
import math
import numpy as np

""" トランジスタ技術 2023年9月号
超安お手軽！ Nucleoのアナログ機能で作る多機能計測ステーション
FFT解析用ユーティリティ

<動作確認環境>
    windows 11
    python 3.8.5

<必要ライブラリ>
    numpy
"""

def analyze_fft(adc_list:list, window_func=None)->tuple:
    """ 時間信号をFFT

    parameters
    -----
    adc_list: list
        ADCの電圧値 0-3.3[V] のリスト
    window_func: function, default None
        窓関数の関数
            hamming => numpy.hamming
            hanning => numpy.hanning
            flattop => scipy.signal.windows.flattop
            None => 方形窓(窓掛けなし)
    
    returns
    -----
    tuple
        2要素のタプル
        1要素目はタプルで、RMS振幅 [Vrms] のFFT結果。実質有効なのは、要素数の半分(ナイキスト周波数)まで。
        2要素目もタプルで、位相 [degree]

        ( (Vrms0, Vrms1, .... ), (Phase0, Phase1, ...) )
          <-- RMS振幅[Vrms] -->   <-- 位相[degree] -->
        
        Vrms0 は 直流電圧 [V]
    """
    num_point = len(adc_list)
    wave_in = np.array(adc_list, dtype=np.float32)
    if window_func:
        wave_in = wave_in * window_func(num_point)
    fft_raw = np.fft.fft(wave_in)
    # 実効値振幅 [Vrms]
    fft_rms = np.abs(fft_raw) / num_point * math.sqrt(2)
    if window_func:
        fft_rms = fft_rms * num_point / sum(window_func(num_point))
    fft_rms[0] = fft_rms[0] / math.sqrt(2)
    # 位相
    fft_degree = np.arctan2(np.imag(fft_raw), np.real(fft_raw)) / np.pi * 180
    return tuple(fft_rms), tuple(fft_degree)

def fft_freq(sampling_freq:float, point:int=8192)->tuple:
    """ FFTの周波数ポイント計算

    parameters
    -----
    sampling_freq: float  ADCサンプリング周波数[Hz]
    point: int, default 8192  ADCサンプリング点数

    returns
    -----
    tuple
        周波数 [Hz] 、実質有効なのは、要素数の半分(ナイキスト周波数)まで。
    """
    return tuple([sampling_freq/point*i for i in range(point)])
